import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from igraph import Graph
from pandas import DataFrame
from shapely import shortest_line


def _od_cost_matrix(
    graph: Graph,
    origins: GeoDataFrame,
    destinations: GeoDataFrame,
    weight: str,
    *,
    lines: bool = False,
    rowwise: bool = False,
) -> DataFrame | GeoDataFrame:
    assert origins.index.name == "temp_idx"
    assert destinations.index.name == "temp_idx"

    results = _get_od_df(graph, origins.index, destinations.index, weight)

    # calculating all-to-all distances is much faster than looping rowwise,
    # so filtering to rowwise afterwards instead
    if rowwise:
        rowwise_df = DataFrame(
            {
                "origin": origins.index,
                "destination": destinations.index,
            }
        )
        results = rowwise_df.merge(results, on=["origin", "destination"], how="left")

    results["wkt_ori"] = results["origin"].map(origins.geometry)
    results["wkt_des"] = results["destination"].map(destinations.geometry)

    results.loc[results.wkt_ori == results.wkt_des, weight] = 0

    # straight lines between origin and destination
    if lines:
        results["geometry"] = shortest_line(results["wkt_ori"], results["wkt_des"])
        results = gpd.GeoDataFrame(results, geometry="geometry", crs=25833)

    results = results.drop(["wkt_ori", "wkt_des"], axis=1, errors="ignore")

    return results.reset_index(drop=True)


def _get_od_df(graph, origins, destinations, weight_col):
    distances: list[list[float]] = graph.distances(
        weights="weight",
        source=origins,
        target=destinations,
    )

    ori_idx, des_idx, costs = [], [], []
    for i, f_idx in enumerate(origins):
        for j, t_idx in enumerate(destinations):
            ori_idx.append(f_idx)
            des_idx.append(t_idx)
            costs.append(distances[i][j])

    return (
        pd.DataFrame(
            data={"origin": ori_idx, "destination": des_idx, weight_col: costs}
        )
        .replace([np.inf, -np.inf], np.nan)
        .reset_index(drop=True)
    )



"""Overlay function that avoids a GEOSException from geopandas.overlay.

This module includes the function 'clean_overlay', which bypasses a
GEOSException from the regular geopandas.overlay. The function is a generalized
version of the solution from GH 2792.

'clean_overlay' also includes the overlay type "update", which can be specified in the
"how" parameter, in addition to the five native geopandas how-s.
"""
import functools

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from pandas import DataFrame
from pyproj import CRS
from shapely import (
    STRtree,
    box,
    difference,
    intersection,
    is_valid,
    make_valid,
    unary_union,
)
from shapely.errors import GEOSException

from .general import clean_geoms
from .geometry_types import get_geom_type, make_all_singlepart, to_single_geom_type


DEFAULT_GRID_SIZE = None
DEFAULT_LSUFFIX = "_1"
DEFAULT_RSUFFIX = "_2"


def clean_overlay(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    how: str = "intersection",
    keep_geom_type: bool = True,
    geom_type: str | None = None,
    grid_size: float | None = None,
    lsuffix: str = DEFAULT_LSUFFIX,
    rsuffix: str = DEFAULT_RSUFFIX,
) -> GeoDataFrame:
    """Fixes and explodes geometries before doing a shapely overlay, then cleans up.

    Fixes geometries, then does a shapely overlay operation that dodges a GEOSException
    raised in the regular geopandas.overlay. The function is a generalised version of a
    solution from GH 2792.

    Args:
        df1: GeoDataFrame
        df2: GeoDataFrame
        how: Method of spatial overlay. Includes the 'update' method, plus the five
            native geopandas methods 'intersection', 'union', 'identity',
            'symmetric_difference' and 'difference'.
        keep_geom_type: If True (default), return only geometries of the same
            geometry type as df1 has, if False, return all resulting geometries.
        geom_type: optionally specify what geometry type to keep before the overlay,
            if there are mixed geometry types. Must be either "polygon", "line" or
            "point".
        grid_size: Precision grid size to round the geometries. Will use the highest
            precision of the inputs by default.

    Returns:
        GeoDataFrame with overlayed and fixed geometries and columns from both
        GeoDataFrames.

    Raises:
        ValueError: If 'how' is not one of 'intersection', 'union', 'identity',
            'symmetric_difference', 'difference' or 'update'.
    """
    # Allowed operations
    allowed_hows = [
        "intersection",
        "union",
        "identity",
        "symmetric_difference",
        "difference",
        "update",
    ]
    if how not in allowed_hows:
        raise ValueError(
            f"`how` was {how!r} but is expected to be in {', '.join(allowed_hows)}"
        )

    if df1.crs != df2.crs:
        raise ValueError(f"'crs' mismatch. Got {df1.crs} and {df2.crs}")

    crs = df1.crs

    original_geom_type = geom_type

    if not geom_type:
        geom_type = get_geom_type(df1)
        if geom_type == "mixed":
            raise ValueError(
                "mixed geometries are not allowed when geom_type isn't specified.",
                df1.geometry,
            )
        if get_geom_type(df2) == "mixed":
            raise ValueError(
                "mixed geometries are not allowed when geom_type isn't specified.",
                df2.geometry,
            )

    df1 = clean_geoms(df1)
    df2 = clean_geoms(df2)

    df1 = make_all_singlepart(df1, ignore_index=True)
    df2 = make_all_singlepart(df2, ignore_index=True)

    df1 = to_single_geom_type(df1, geom_type)

    if original_geom_type:
        df2 = to_single_geom_type(df2, geom_type)

    assert df1.is_valid.all()
    assert df2.is_valid.all()
    assert df1.geometry.notna().all()
    assert df2.geometry.notna().all()

    box1 = box(*df1.total_bounds)
    box2 = box(*df2.total_bounds)

    if not len(df1) or not len(df1) or not box1.intersects(box2):
        return _no_intersections_return(df1, df2, how, lsuffix, rsuffix)

    if df1._geometry_column_name != "geometry":
        df1 = df1.rename_geometry("geometry")

    if df2._geometry_column_name != "geometry":
        df2 = df2.rename_geometry("geometry")

    # to pandas because GeoDataFrame constructor is expensive
    df1 = DataFrame(df1).reset_index(drop=True)
    df2 = DataFrame(df2).reset_index(drop=True)

    overlayed = (
        gpd.GeoDataFrame(
            _shapely_pd_overlay(
                df1,
                df2,
                how=how,
                grid_size=grid_size,
                lsuffix=lsuffix,
                rsuffix=rsuffix,
            ),
            geometry="geometry",
            crs=crs,
        )
        .pipe(clean_geoms)
        .pipe(make_all_singlepart, ignore_index=True)
    )

    if keep_geom_type:
        overlayed = to_single_geom_type(overlayed, geom_type)

    return overlayed.reset_index(drop=True)


def _join_and_get_no_rows(df1, df2, lsuffix, rsuffix):
    geom_col = df1._geometry_column_name
    df1_cols = df1.columns.difference({geom_col})
    df2_cols = df2.columns.difference({df2._geometry_column_name})
    cols_with_suffix = [
        f"{col}{lsuffix}" if col in df2_cols else col for col in df1_cols
    ] + [f"{col}{rsuffix}" if col in df1_cols else col for col in df2_cols]

    return GeoDataFrame(
        pd.DataFrame(columns=cols_with_suffix + [geom_col]),
        geometry=geom_col,
        crs=df1.crs,
    )


def _no_intersections_return(df1, df2, how, lsuffix, rsuffix):
    """Return with no overlay if no intersecting bounding box"""

    if how == "intersection":
        return _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)

    if how == "difference":
        return df1.reset_index(drop=True)

    if how == "identity":
        # add suffixes and return df1
        df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
        df2_cols = df2.columns.difference({df2._geometry_column_name})
        df1.columns = [f"{col}{lsuffix}" if col in df2_cols else col for col in df1]
        return pd.concat([df_template, df1], ignore_index=True)

    if how == "update":
        return pd.concat([df1, df2], ignore_index=True)

    assert how in ["union", "symmetric_difference"]

    # add suffixes and return both concatted
    df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
    if not len(df1) and not len(df2):
        return df_template

    df_template = _join_and_get_no_rows(df1, df2, lsuffix, rsuffix)
    return pd.concat([df_template, df1, df2], ignore_index=True)


def _shapely_pd_overlay(
    df1: DataFrame,
    df2: DataFrame,
    how: str,
    grid_size: float = DEFAULT_GRID_SIZE,
    lsuffix=DEFAULT_LSUFFIX,
    rsuffix=DEFAULT_RSUFFIX,
) -> DataFrame:
    if not grid_size and not len(df1) or not len(df2):
        return _no_intersections_return(df1, df2, how, lsuffix, rsuffix)

    tree = STRtree(df2.geometry.values)
    left, right = tree.query(df1.geometry.values, predicate="intersects")

    pairs = _get_intersects_pairs(df1, df2, left, right, rsuffix)
    assert pairs.geometry.notna().all()
    assert pairs.geom_right.notna().all()

    if how == "intersection":
        overlayed = [_intersection(pairs, grid_size=grid_size)]

    elif how == "difference":
        overlayed = _difference(pairs, df1, left, grid_size=grid_size)

    elif how == "symmetric_difference":
        overlayed = _symmetric_difference(
            pairs, df1, df2, left, right, grid_size=grid_size, rsuffix=rsuffix
        )

    elif how == "identity":
        overlayed = _identity(pairs, df1, left, grid_size=grid_size)

    elif how == "union":
        overlayed = _union(
            pairs, df1, df2, left, right, grid_size=grid_size, rsuffix=rsuffix
        )

    elif how == "update":
        overlayed = _update(pairs, df1, df2, left=left, grid_size=grid_size)

    assert isinstance(overlayed, list)

    overlayed = pd.concat(overlayed, ignore_index=True).drop(
        columns="_overlay_index_right", errors="ignore"
    )

    # push geometry column to the end
    overlayed = overlayed.reindex(
        columns=[c for c in overlayed.columns if c != "geometry"] + ["geometry"]
    )

    if how not in ["difference", "update"]:
        overlayed = _add_suffix_left(overlayed, df1, df2, lsuffix)

    overlayed["geometry"] = make_valid(overlayed["geometry"])
    # None and empty are falsy
    overlayed = overlayed.loc[lambda x: x["geometry"].notna()]

    return overlayed


def _update(pairs, df1, df2, left, grid_size) -> GeoDataFrame:
    overlayed = _difference(pairs, df1, left, grid_size=grid_size)

    return overlayed + [df2]


def _intersection(pairs, grid_size) -> GeoDataFrame:
    if not len(pairs):
        return pairs.drop(columns="geom_right")

    # assert all(is_valid(pairs["geometry"].to_numpy())), "\nnot all valid\n"
    # assert all(is_valid(pairs["geom_right"].to_numpy())), "\nnot all valid\n"

    intersections = pairs.copy()
    """
    geoms_left = make_valid(intersections["geometry"].to_numpy())
    geoms_right = make_valid(intersections["geom_right"].to_numpy())
    intersections["geometry"] = intersection(
        geoms_left,
        geoms_right,
        grid_size=grid_size,
    )
    try:
        intersections["geometry"] = intersection(
            intersections["geometry"].to_numpy(),
            intersections["geom_right"].to_numpy(),
            grid_size=grid_size,
        )
    except GEOSException:"""
    intersections["geometry"] = intersection(
        intersections["geometry"].to_numpy(),
        intersections["geom_right"].to_numpy(),
        grid_size=grid_size,
    )

    return intersections.drop(columns="geom_right")


def _union(pairs, df1, df2, left, right, grid_size, rsuffix):
    merged = []
    if len(left):
        intersections = _intersection(pairs, grid_size=grid_size)
        merged.append(intersections)
    symmdiff = _symmetric_difference(
        pairs, df1, df2, left, right, grid_size=grid_size, rsuffix=rsuffix
    )
    merged += symmdiff
    return merged


def _identity(pairs, df1, left, grid_size):
    merged = []
    if len(left):
        intersections = _intersection(pairs, grid_size=grid_size)
        merged.append(intersections)
    diff = _difference(pairs, df1, left, grid_size=grid_size)
    merged += diff
    return merged


def _symmetric_difference(pairs, df1, df2, left, right, grid_size, rsuffix) -> list:
    merged = []

    difference_left = _difference(pairs, df1, left, grid_size=grid_size)
    merged += difference_left

    if len(left):
        clip_right = _shapely_diffclip_right(
            pairs, df1, df2, grid_size=grid_size, rsuffix=rsuffix
        )
        merged.append(clip_right)

    diff_right = _add_from_right(df1, df2, right, rsuffix)
    merged.append(diff_right)

    return merged


def _difference(pairs, df1, left, grid_size=None) -> list:
    merged = []
    if len(left):
        clip_left = _shapely_diffclip_left(pairs=pairs, df1=df1, grid_size=grid_size)
        merged.append(clip_left)
    diff_left = _add_indices_from_left(df1, left)
    merged.append(diff_left)
    return merged


def _get_intersects_pairs(
    df1: GeoDataFrame,
    df2: GeoDataFrame,
    left: np.ndarray,
    right: np.ndarray,
    rsuffix,
) -> DataFrame:
    return pd.concat(
        [
            df1.take(left),
            (
                DataFrame(
                    {"_overlay_index_right": right}, index=df1.index.values.take(left)
                )
            ),
        ],
        axis=1,
    ).join(
        df2.rename(columns={"geometry": "geom_right"}, errors="raise"),
        on="_overlay_index_right",
        rsuffix=rsuffix,
    )


def _add_suffix_left(overlayed, df1, df2, lsuffix):
    """Separating this from _add_indices_from_left, since this suffix is not needed in difference."""
    return overlayed.rename(
        columns={
            c: f"{c}{lsuffix}"
            if c in df1.columns and c in df2.columns and c != "geometry"
            else c
            for c in overlayed.columns
        }
    )


def _add_indices_from_left(df1, left):
    return df1.take(np.setdiff1d(np.arange(len(df1)), left))


def _add_from_right(
    df1: GeoDataFrame, df2: GeoDataFrame, right: np.ndarray, rsuffix
) -> GeoDataFrame:
    return df2.take(np.setdiff1d(np.arange(len(df2)), right)).rename(
        columns={
            c: f"{c}{rsuffix}" if c in df1.columns and c != "geometry" else c
            for c in df2.columns
        }
    )


def _shapely_diffclip_left(pairs, df1, grid_size):
    """Aggregate areas in right by unique values of left, then use those to clip
    areas out of left"""

    agg_geoms_partial = functools.partial(agg_geoms, grid_size=grid_size)

    clip_left = pairs.groupby(level=0).agg(
        {
            "geom_right": agg_geoms_partial,
            **{
                c: "first"
                for c in df1.columns
                if c not in ["_overlay_index_right", "geom_right"]
            },
        }
    )

    assert clip_left["geometry"].notna().all()
    assert clip_left["geom_right"].notna().all()

    clip_left["geometry"] = _try_difference(
        clip_left["geometry"].to_numpy(),
        clip_left["geom_right"].to_numpy(),
        grid_size=grid_size,
    )

    return clip_left.drop(columns="geom_right")


def _shapely_diffclip_right(pairs, df1, df2, grid_size, rsuffix):
    agg_geoms_partial = functools.partial(agg_geoms, grid_size=grid_size)

    clip_right = (
        pairs.rename(columns={"geometry": "geom_left", "geom_right": "geometry"})
        .groupby(by="_overlay_index_right")
        .agg(
            {
                "geom_left": agg_geoms_partial,
                "geometry": "first",
            }
        )
        .join(df2.drop(columns=["geometry"]))
        .rename(
            columns={
                c: f"{c}{rsuffix}" if c in df1.columns and c != "geometry" else c
                for c in df2.columns
            }
        )
    )

    assert clip_right["geometry"].notna().all()
    assert clip_right["geom_left"].notna().all()

    clip_right["geometry"] = _try_difference(
        clip_right["geometry"].to_numpy(),
        clip_right["geom_left"].to_numpy(),
        grid_size=grid_size,
    )

    return clip_right.drop(columns="geom_left")


def _try_difference(left, right, grid_size):
    """Try difference overlay, then make_valid and retry."""
    try:
        return difference(
            left,
            right,
            grid_size=grid_size,
        )
    except GEOSException:
        return difference(
            make_valid(left),
            make_valid(right),
            grid_size=grid_size,
        )


def agg_geoms(g, grid_size=None):
    return (
        make_valid(unary_union(g, grid_size=grid_size)) if len(g) > 1 else make_valid(g)
    )



import numbers
import re
import uuid
import warnings
from collections.abc import Callable
from copy import copy, deepcopy
from json import loads
from pathlib import Path

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
import rasterio
import shapely


try:
    import xarray as xr
except ImportError:
    pass
try:
    from rioxarray.rioxarray import _generate_spatial_coords
except ImportError:
    pass
from affine import Affine
from geopandas import GeoDataFrame, GeoSeries
from pandas.api.types import is_list_like
from rasterio import features
from rasterio.enums import MergeAlg
from rasterio.io import DatasetReader
from rasterio.mask import mask as rast_mask
from rasterio.vrt import WarpedVRT
from rasterio.warp import reproject
from shapely import Geometry, box
from shapely.geometry import Point, Polygon, shape

from ..geopandas_tools.bounds import to_bbox
from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.general import is_bbox_like, is_wkt
from ..helpers import get_non_numpy_func_name, get_numpy_func, is_property
from ..io.opener import opener
from .base import (
    RasterBase,
    RasterHasChangedError,
    get_index_mapper,
    memfile_from_array,
)
from .zonal import (
    _aggregate,
    _no_overlap_df,
    make_geometry_iterrows,
    prepare_zonal,
    zonal_post,
)


numpy_func_message = (
    "aggfunc must be functions or " "strings of numpy functions or methods."
)


class Raster(RasterBase):
    """For reading, writing and working with rasters.

    Raster instances should be created with the methods 'from_path', 'from_array' or
    'from_gdf'.


    Examples
    --------

    Read tif file.

    >>> path = 'https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif'
    >>> raster = sg.Raster.from_path(path)
    >>> raster
    Raster(shape=(1, 201, 201), res=10, crs=ETRS89 / UTM zone 33N (N-E), path=https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif)

    Load the entire image as an numpy ndarray.
    Operations are done in place to save memory.
    The array is stored in the array attribute.

    >>> raster.load()
    >>> raster.array[raster.array < 0] = 0
    >>> raster.array
    [[[  0.    0.    0.  ... 158.4 155.6 152.6]
    [  0.    0.    0.  ... 158.  154.8 151.9]
    [  0.    0.    0.  ... 158.5 155.1 152.3]
    ...
    [  0.  150.2 150.6 ...   0.    0.    0. ]
    [  0.  149.9 150.1 ...   0.    0.    0. ]
    [  0.  149.2 149.5 ...   0.    0.    0. ]]]

    Save as tif file.

    >>> raster.write("path/to/file.tif")

    Convert to GeoDataFrame.

    >>> gdf = raster.to_gdf(column="elevation")
    >>> gdf
           elevation                                           geometry  band_index
    0            1.9  POLYGON ((-25665.000 6676005.000, -25665.000 6...           1
    1           11.0  POLYGON ((-25655.000 6676005.000, -25655.000 6...           1
    2           18.1  POLYGON ((-25645.000 6676005.000, -25645.000 6...           1
    3           15.8  POLYGON ((-25635.000 6676005.000, -25635.000 6...           1
    4           11.6  POLYGON ((-25625.000 6676005.000, -25625.000 6...           1
    ...          ...                                                ...         ...
    25096       13.4  POLYGON ((-24935.000 6674005.000, -24935.000 6...           1
    25097        9.4  POLYGON ((-24925.000 6674005.000, -24925.000 6...           1
    25098        5.3  POLYGON ((-24915.000 6674005.000, -24915.000 6...           1
    25099        2.3  POLYGON ((-24905.000 6674005.000, -24905.000 6...           1
    25100        0.1  POLYGON ((-24895.000 6674005.000, -24895.000 6...           1

    The image can also be clipped by a mask while loading.

    >>> small_circle = raster_as_polygons.unary_union.centroid.buffer(50)
    >>> raster = sg.Raster.from_path(path).clip(small_circle)
    Raster(shape=(1, 10, 10), res=10, crs=ETRS89 / UTM zone 33N (N-E), path=https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/raster/dtm_10.tif)

    Construct raster from GeoDataFrame.
    The arrays are put on top of each other in a 3 dimensional array.

    >>> raster_as_polygons["elevation_x2"] = raster_as_polygons["elevation"] * 2
    >>> raster_from_polygons = sg.Raster.from_gdf(raster_as_polygons, columns=["elevation", "elevation_x2"], res=20)
    >>> raster_from_polygons
    Raster(shape=(2, 100, 100), res=20, raster_id=-260056673995, crs=ETRS89 / UTM zone 33N (N-E), path=None)

    Calculate zonal statistics for each polygon in 'gdf'.

    >>> zonal = raster.zonal(raster_as_polygons, aggfunc=["sum", np.mean])
    >>> zonal
            sum  mean                                           geometry
    0       1.9   1.9  POLYGON ((-25665.000 6676005.000, -25665.000 6...
    1      11.0  11.0  POLYGON ((-25655.000 6676005.000, -25655.000 6...
    2      18.1  18.1  POLYGON ((-25645.000 6676005.000, -25645.000 6...
    3      15.8  15.8  POLYGON ((-25635.000 6676005.000, -25635.000 6...
    4      11.6  11.6  POLYGON ((-25625.000 6676005.000, -25625.000 6...
    ...     ...   ...                                                ...
    25096  13.4  13.4  POLYGON ((-24935.000 6674005.000, -24935.000 6...
    25097   9.4   9.4  POLYGON ((-24925.000 6674005.000, -24925.000 6...
    25098   5.3   5.3  POLYGON ((-24915.000 6674005.000, -24915.000 6...
    25099   2.3   2.3  POLYGON ((-24905.000 6674005.000, -24905.000 6...
    25100   0.1   0.1  POLYGON ((-24895.000 6674005.000, -24895.000 6...

    """

    def __init__(
        self,
        raster=None,
        *,
        path: str | None = None,
        array: np.ndarray | None = None,
        band_index: int | list[int] | None = None,
        name: str | None = None,
        name_regex: str | None = None,
        date: str | None = None,
        date_regex: str | None = None,
        shortname: str | None = None,
        dtype: type | None = None,
        nodata: int | float | None = None,
        **kwargs,
    ):
        if raster is not None:
            if not isinstance(raster, Raster):
                raise TypeError(
                    "Raster should be constructed with the from-classmethods."
                )
            for key, value in raster.__dict__.items():
                self[key] = value
            return

        self.path = path
        self.array = array

        # hasattr check to allow class attributes in subclasses
        if not hasattr(self, "shortname"):
            self.shortname = shortname
        if not hasattr(self, "name_regex"):
            self.name_regex = name_regex
        if not hasattr(self, "date_regex"):
            self.date_regex = date_regex
        if not hasattr(self, "_dtype") or dtype is not None:
            self._dtype = dtype
        if not hasattr(self, "nodata"):
            self.nodata = nodata

        if not hasattr(self, "driver"):
            self.driver = kwargs.pop("driver", "GTiff")
        if not hasattr(self, "compress"):
            self.compress = kwargs.pop("compress", "LZW")

        if not hasattr(self, "_band_index") or band_index is not None:
            self._band_index = self._get_band_index(band_index)

        # underscore to allow properties without setters in subclasses (e.g. Sentinel2)
        if date:
            self._date = date
        if name:
            self._name = name

        self.root = kwargs.pop("root", None)

        self._hash = uuid.uuid4()

        # override the above with kwargs
        self.update(**kwargs)

        if self.path is None and self.array is None:
            return

        if self.path is not None and self.array is not None:
            raise ValueError("Cannot supply both 'path' and 'array'.")

        # the important attributes must be of correct type
        crs = kwargs.get("crs")
        transform = kwargs.get("transform")
        bounds = kwargs.get("bounds")
        if bounds is not None:
            bounds = to_bbox(bounds)

        if not any([path, transform, bounds]):
            raise TypeError(
                "Must specify either bounds or transform when constructing raster from array."
            )

        self._crs = pyproj.CRS(crs) if crs else None
        transform = Affine(*transform) if transform is not None else None

        if transform and not bounds:
            self._bounds = rasterio.transform.array_bounds(
                self.height, self.width, transform
            )
        elif not path:
            self._bounds = bounds

        attributes = set(self.__dict__.keys()).difference(set(self.properties))

        if self.path is not None and not self.has_nessecary_attrs(attributes):
            self.add_meta()
            self._meta_added = True

        if self.path is None:
            if not isinstance(self.array, np.ndarray):
                raise TypeError

            if not hasattr(self, "crs"):
                raise TypeError("Must specify crs when constructing raster from array.")

            self._band_index = self._add_band_index_from_array(band_index)

        self._prev_crs = self._crs

    @classmethod
    def from_path(
        cls,
        path: str,
        **kwargs,
    ):
        """Construct Raster from file path.

        Args:
            path: Path to a raster image file.

        Returns:
            A Raster instance.
        """
        return cls(
            path=str(path),
            **kwargs,
        )

    @classmethod
    def from_array(
        cls,
        array: np.ndarray,
        *,
        crs=None,
        transform: Affine | None = None,
        bounds: tuple | Geometry | None = None,
        copy: bool = True,
        **kwargs,
    ):
        """Construct Raster from numpy array.

        Metadata must be specified, either individually or in a dictionary
        (hint: use the 'meta' attribute of an existing Raster object if applicable).
        The necessary metadata is 'crs' and either 'transform' (Affine object) or 'bounds',
        which transform will then be created from.

        Args:
            array: 2d or 3d numpy ndarray.
            crs: Coordinate reference system.
            transform: Affine transform object. Can be specified instead
                of bounds.
            bounds: Minimum and maximum x and y coordinates. Can be specified instead
                of transform.
            name: Optional name to give the raster.

        Returns:
            A Raster instance.
        """
        if array is None:
            raise TypeError("Must specify array.")

        array = array.copy() if copy else array

        return cls(
            array=array,
            crs=crs,
            transform=transform,
            bounds=bounds,
            **kwargs,
        )

    @classmethod
    def from_gdf(
        cls,
        gdf: GeoDataFrame,
        columns: str | list[str],
        res: int,
        fill=0,
        all_touched=False,
        merge_alg=MergeAlg.replace,
        default_value=1,
        dtype=None,
        **kwargs,
    ):
        """Construct Raster from a GeoDataFrame.

        Args:
            gdf: The GeoDataFrame.
            column: The column to be used as values for the array.
            res: Resolution of the raster in units of gdf's coordinate
                reference system.

        """
        if not isinstance(gdf, GeoDataFrame):
            gdf = to_gdf(gdf)

        if "transform" in kwargs:
            raise TypeError("Unexpected argument 'transform'")

        kwargs["crs"] = gdf.crs or kwargs.get("crs")

        if kwargs["crs"] is None:
            raise TypeError("Must specify crs if the object doesn't have crs.")

        shape = cls.get_shape_from_bounds(gdf.total_bounds, res=res)
        transform = cls.get_transform_from_bounds(gdf.total_bounds, shape)
        kwargs["transform"] = transform

        def _rasterize(gdf, col):
            return features.rasterize(
                cls._to_geojson_geom_val(gdf, col),
                out_shape=shape,
                transform=transform,
                fill=fill,
                all_touched=all_touched,
                merge_alg=merge_alg,
                default_value=default_value,
                dtype=dtype,
            )

        # make 2d array
        if isinstance(columns, str):
            array = _rasterize(gdf, columns)
            assert len(array.shape) == 2
            name = kwargs.get("name", columns)

        # 3d array even if single column in list/tuple
        elif hasattr(columns, "__iter__"):
            array = []
            for col in columns:
                arr = _rasterize(gdf, col)
                array.append(arr)
            array = np.array(array)
            assert len(array.shape) == 3
            name = kwargs.get("name", None)

        return cls.from_array(array=array, name=name, **kwargs)

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Construct Raster from metadata dict to fastpass the initializer.

        This is the fastest way to create a Raster since a metadata lookup is not
        needed.

        The dictionary must have all the keys ...
        and at least one of the keys 'transform' and 'bounds'.

        Args:
            dictionary: Dictionary with the nessecary and optional information
                about the raster. This can be fetched from an existing raster with
                the to_dict method.

        Returns:
            A Raster instance.
        """

        cls.validate_dict(dictionary)

        return cls(**dictionary)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            self.validate_key(key)
            if key == "indexes":
                self._band_index = value
            if is_property(self, key):
                self["_" + key] = value
            else:
                self[key] = value
        return self

    def load(self, res: int | None = None, **kwargs):
        """Load the entire image as an np.array.

        The array is stored in the 'array' attribute
        of the Raster.

        Args:
            **kwargs: Keyword arguments passed to the rasterio read
                method.
        """
        if "mask" in kwargs:
            raise ValueError("Got an unexpected keyword argument 'mask'")
        if "window" in kwargs:
            raise ValueError("Got an unexpected keyword argument 'window'")

        self._read_tif(res=res, **kwargs)

        return self

    def clip(self, mask, boundless: bool = True, **kwargs):
        """Load the part of the image inside the mask.

        The returned array is stored in the 'array' attribute
        of the Raster.

        Args:
            mask: Geometry-like object or bounding box.
            **kwargs: Keyword arguments passed to the mask function
                from the rasterio.mask module.

        Returns:
            Self, but with the array loaded.
        """
        if not isinstance(mask, GeoDataFrame):
            mask = self._return_gdf(mask)

        if not self.crs.equals(pyproj.CRS(mask.crs)):
            raise ValueError("crs mismatch.")

        self._read_with_mask(mask=mask, boundless=boundless, **kwargs)

        return self

    def sample(self, n=1, size=20, mask=None, copy=True, **kwargs):
        if mask is not None:
            points = GeoSeries(self.unary_union).clip(mask).sample_points(n)
        else:
            points = GeoSeries(self.unary_union).sample_points(n)
        buffered = points.buffer(size / self.res[0])
        boxes = to_gdf(
            [shapely.box(*arr) for arr in buffered.bounds.values], crs=self.crs
        )
        if copy:
            copy = self.copy()
            return copy.clip(boxes, **kwargs)
        return self.clip(boxes, **kwargs)

    def zonal(
        self,
        polygons: GeoDataFrame,
        aggfunc: str | Callable | list[Callable | str],
        array_func: Callable | None = None,
        dropna: bool = True,
    ) -> GeoDataFrame:
        """Calculate zonal statistics in polygons.

        Args:
            polygons: A GeoDataFrame of polygon geometries.
            aggfunc: Function(s) of which to aggregate the values
                within each polygon.
            array_func: Optional calculation of the raster
                array before calculating the zonal statistics.
            dropna: If True (default), polygons with all missing
                values will be removed.

        Returns:
            A GeoDataFrame with aggregated values per polygon.
        """
        idx_mapper, idx_name = get_index_mapper(polygons)
        polygons, aggfunc, func_names = prepare_zonal(polygons, aggfunc)
        poly_iter = make_geometry_iterrows(polygons)

        aggregated = []
        for i, poly in poly_iter:
            clipped = self.clip(poly)
            if not np.size(clipped.array):
                aggregated.append(_no_overlap_df(func_names, i, date=self.date))
            aggregated.append(
                _aggregate(clipped.array, array_func, aggfunc, func_names, self.date, i)
            )

        return zonal_post(
            aggregated,
            polygons=polygons,
            idx_mapper=idx_mapper,
            idx_name=idx_name,
            dropna=dropna,
        )

    def write(self, path: str, window=None, **kwargs):
        """Write the raster as a single file.

        Multiband arrays will result in a multiband image file.

        Args:
            path: File path to write to.
            window: Optional window to clip the image to.
        """

        if self.array is None:
            raise AttributeError("The image hasn't been loaded yet.")

        profile = self.profile | kwargs

        with opener(path) as file:
            with rasterio.open(file, "w", **profile) as dst:
                self._write(dst, window)

        self.path = str(path)

    def to_xarray(self) -> xr.DataArray:
        self.check_for_array()
        coords = _generate_spatial_coords(self.transform, self.width, self.height)
        if len(self.array.shape) == 2:
            dims = ["y", "x"]
        elif len(self.array.shape) == 3:
            dims = ["band", "y", "x"]
        else:
            raise ValueError("Array must be 2 or 3 dimensional.")
        return xr.DataArray(
            self.array,
            coords=coords,
            dims=dims,
            name=self.name,
            attrs={"crs": self.crs},
        )  # .transpose("y", "x")

    def to_dict(self) -> dict:
        out = {}
        for col in self.ALL_ATTRS:
            try:
                out[col] = self[col]
            except AttributeError:
                pass
        return out

    def to_gdf(self, column: str | list[str] | None = None) -> GeoDataFrame:
        """Create a GeoDataFrame from the raster.

        For multiband rasters, the bands are in separate rows with a "band" column
        value corresponding to the band indexes of the raster.

        Args:
            column: Name of resulting column(s) that holds the raster values.
                Can be a single string or an iterable with the same length as
                the number of raster bands.

        Returns:
            A GeoDataFrame with a geometry column, a 'band' column and a
            one or more value columns.
        """
        self.check_for_array()

        array_list = self._to_2d_array_list(self.array)

        if is_list_like(column) and len(column) != len(array_list):
            raise ValueError(
                "columns should be a string or a list of same length as "
                f"layers in the array ({len(array_list)})."
            )

        if column is None:
            column = ["value"] * len(array_list)

        if isinstance(column, str):
            column = [column] * len(array_list)

        gdfs = []
        for i, (column, array) in enumerate(zip(column, array_list, strict=True)):
            gdf = gpd.GeoDataFrame(
                pd.DataFrame(
                    self._array_to_geojson(array, self.transform),
                    columns=[column, "geometry"],
                ),
                geometry="geometry",
                crs=self.crs,
            )
            gdf["band_index"] = i + 1

            if hasattr(self, "_datadict"):
                for name, value in self._datadict.items():
                    try:
                        gdf[name] = value
                    except Exception:
                        # in case of iterable columns (band_index...)
                        # reset index in case of duplicate index
                        index = gdf.index
                        gdf = gdf.reset_index(drop=True)
                        gdf[name] = pd.Series(
                            [value for _ in range(len(gdf))], index=gdf.index
                        )
                        gdf.index = index

            gdfs.append(gdf)

        return pd.concat(gdfs, ignore_index=True)

    def set_crs(
        self,
        crs,
        allow_override: bool = False,
    ):
        """Set coordinate reference system."""
        if not allow_override and self.crs is not None:
            raise ValueError("Cannot overwrite crs when allow_override is False.")

        if self.array is None:
            raise ValueError("array must be loaded/clipped before set_crs")

        self._crs = pyproj.CRS(crs)
        return self

    def to_crs(self, crs, **kwargs):
        """Reproject the raster.

        Args:
            crs: The new coordinate reference system.
            **kwargs: Keyword arguments passed to the reproject function
                from the rasterio.warp module.
        """
        if self.crs is None:
            raise ValueError("Raster has no crs. Use set_crs.")

        if pyproj.CRS(crs).equals(pyproj.CRS(self.crs)):
            return self

        if self.array is None:
            project = pyproj.Transformer.from_crs(
                pyproj.CRS(self._prev_crs), pyproj.CRS(crs), always_xy=True
            ).transform

            old_box = shapely.box(*self.bounds)
            new_box = shapely.ops.transform(project, old_box)
            self._bounds = to_bbox(new_box)

            # TODO: fix this
            print("old/new:", shapely.area(old_box) / shapely.area(new_box))

            """self._bounds = rasterio.warp.transform_bounds(
                pyproj.CRS(self._prev_crs), pyproj.CRS(crs), *to_bbox(self._bounds)
            )
            transformer = pyproj.Transformer.from_crs(
                pyproj.CRS(self._prev_crs), pyproj.CRS(crs), always_xy=True
            )
            minx, miny, maxx, maxy = self.bounds
            xs, ys = transformer.transform(xx=[minx, maxx], yy=[miny, maxy])

            minx, maxx = xs
            miny, maxy = ys
            self._bounds = minx, miny, maxx, maxy"""

            # self._bounds = shapely.transform(old_box, project)
        else:
            self.array, transform = reproject(
                source=self.array,
                src_crs=self.crs,
                src_transform=self.transform,
                dst_crs=pyproj.CRS(crs),
                **kwargs,
            )

            self._bounds = rasterio.transform.array_bounds(
                self.height, self.width, transform
            )

        self._warped_crs = pyproj.CRS(crs)
        self._prev_crs = pyproj.CRS(crs)

        return self

    def plot(self, mask=None) -> None:
        self.check_for_array()
        """Plot the images. One image per band."""
        if len(self.shape) == 3:
            for arr in self.array:
                self._plot_2d(arr)

    def astype(self, dtype: type):
        if self.array is None:
            raise ValueError("Array is not loaded.")
        if not rasterio.dtypes.can_cast_dtype(self.array, dtype):
            min_dtype = rasterio.dtypes.get_minimum_dtype(self.array)
            raise ValueError(f"Cannot cast to dtype. Minimum dtype is {min_dtype}")
        self.array = self.array.astype(dtype)
        self._dtype = dtype
        return self

    def as_minimum_dtype(self):
        min_dtype = rasterio.dtypes.get_minimum_dtype(self.array)
        self.array = self.array.astype(min_dtype)
        return self

    def add_meta(self):
        mess = "Cannot add metadata after image has been "
        if hasattr(self, "_clipped"):
            raise ValueError(mess + "clipped.")
        if hasattr(self, "_warped_crs"):
            raise ValueError(mess + "reprojected.")

        with opener(self.path) as file:
            with rasterio.open(file) as src:
                self._add_meta_from_src(src)

        return self

    def get_coords(self):
        # TODO: droppe
        self.check_for_array()
        return _generate_spatial_coords(self.transform, self.width, self.height)

    @property
    def subfolder(self):
        try:
            folder = Path(Path(self.path).parent).name
            if self.root:
                return folder.difference(self.root)
            else:
                return folder
        except TypeError:
            return None

    @property
    def tile(self) -> str | None:
        if self.bounds is None:
            return None
        return f"{int(self.bounds[0])}{int(self.bounds[1])}"

    @property
    def raster_id(self) -> str:
        shortname = self.shortname if self.shortname else ""
        date = self.date if self.date else ""
        name = self.name if self.name else ""
        return f"{shortname}_{self.tile}_{date}_{name}".replace("__", "_").strip("_")

    @property
    def band_index(self):
        return self._band_index

    @property
    def name(self) -> str | None:
        if hasattr(self, "_name"):
            return self._name
        if not self.name_regex and self.path:
            return Path(self.path).stem
        try:
            return re.search(self.name_regex, Path(self.path).name).group()
        except (AttributeError, TypeError):
            return None

    @name.setter
    def name(self, value):
        self._name = value
        return self._name

    @property
    def date(self):
        if hasattr(self, "_date"):
            return self._date
        try:
            return re.search(self.date_regex, Path(self.path).name).group()
        except (AttributeError, TypeError):
            return None

    @date.setter
    def date(self, value):
        self._date = value
        return self._date

    @property
    def band_color(self):
        """To be implemented in subclasses."""
        pass

    @property
    def is_mask(self):
        """To be implemented in subclasses."""
        pass

    def array_list(self):
        return self._to_2d_array_list(self.array)

    @property
    def dtype(self):
        try:
            return self.array.dtype
        except Exception:
            return getattr(self, "_dtype", None)

    @dtype.setter
    def dtype(self, new_dtype):
        self.array = self.array.astype(new_dtype)
        return self.array.dtype

    def read_kwargs(self, kwargs):
        # if kwargs is None:
        #   kwargs = {}
        return {
            "indexes": self.band_index,
            "masked": False,
            "fill_value": self.nodata,
        } | kwargs or {}

    @property
    def profile(self):
        # TODO: .crs blir feil hvis warpa. Eller?
        return {
            "driver": self.driver,
            "dtype": self.dtype,
            "crs": self.crs,
            "transform": self.transform,
            "nodata": self.nodata,
            "count": self.count,
            "height": self.height,
            "width": self.width,
            "compress": self.compress,
            "indexes": self.band_index,
        }

    @property
    def raster_type(self) -> str:
        return self.__class__.__name__

    @property
    def area(self):
        return shapely.area(self.unary_union)

    @property
    def length(self):
        return shapely.length(self.unary_union)

    @property
    def crs(self):
        try:
            return self._warped_crs
        except AttributeError:
            try:
                return self._crs
            except AttributeError:
                return None

    @property
    def unary_union(self) -> Polygon:
        return shapely.box(*self.bounds)

    @property
    def centroid(self) -> Point:
        x = (self.bounds[0] + self.bounds[2]) / 2
        y = (self.bounds[1] + self.bounds[3]) / 2
        return Point(x, y)

    @property
    def res(self) -> tuple[float, float]:
        if self.width is None:
            return None
        diffx = self.bounds[2] - self.bounds[0]
        diffy = self.bounds[3] - self.bounds[1]
        resx = diffx / self.width
        resy = diffy / self.height
        return resx, resy

    @property
    def height(self):
        if self.array is None:
            try:
                return self._height
            except AttributeError:
                return None
        i = 1 if len(self.array.shape) == 3 else 0
        return self.array.shape[i]

    @property
    def width(self):
        if self.array is None:
            try:
                return self._width
            except AttributeError:
                return None
        i = 2 if len(self.array.shape) == 3 else 1
        return self.array.shape[i]

    @property
    def count(self):
        if self.array is not None:
            if len(self.array.shape) == 3:
                return self.array.shape[0]
            if len(self.array.shape) == 2:
                return 1
        if not hasattr(self._band_index, "__iter__"):
            return 1
        return len(self._band_index)

    @property
    def shape(self):
        """Shape that is consistent with the array, whether it is loaded or not."""
        if self.array is not None:
            return self.array.shape
        if self._band_index is None or hasattr(self._band_index, "__iter__"):
            return self.count, self.width, self.height
        return self.width, self.height

    @property
    def transform(self) -> Affine:
        return rasterio.transform.from_bounds(*self.bounds, self.width, self.height)

    @property
    def bounds(self) -> tuple[float, float, float, float] | None:
        if not hasattr(self, "_bounds"):
            return None
        return to_bbox(self._bounds)

    @property
    def total_bounds(self) -> tuple[float, float, float, float]:
        return self.bounds

    @classmethod
    def has_nessecary_attrs(cls, dict_like):
        """Check if Raster init got enough kwargs to not need to read src."""
        try:
            cls.validate_dict(dict_like)
            return True
        except AttributeError:
            return False

    def band_index_tuple(self) -> tuple[int, ...]:
        if len(self.shape) == 2:
            return (1,)
        return tuple(i + 1 for i in range(self.shape[0]))

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, ValueError, IndexError):
            return default

    def copy(self, deep=True):
        """Returns a (deep) copy of the class instance.

        Args:
            deep: Whether to return a deep or shallow copy. Defaults to True.
        """
        if deep:
            return deepcopy(self)
        else:
            return copy(self)

    def equals(self, other) -> bool:
        if not isinstance(other, Raster):
            raise NotImplementedError("other must be of type Raster")
        if self.array is None and other.array is not None:
            return False
        if self.array is not None and other.array is None:
            return False

        for method in dir(self):
            if not is_property(self, method):
                continue
            if self[method] != other[method]:
                return False

        return np.array_equal(self.array, other.array)

    @staticmethod
    def _plot_2d(array, mask=None) -> None:
        ax = plt.axes()
        ax.imshow(array)
        ax.axis("off")
        plt.show()
        plt.close()

    def check_for_array(self, text=""):
        mess = "Arrays are not loaded. " + text
        if self.array is None:
            raise ValueError(mess)

    def _add_band_index_from_array(self, band_index):
        if band_index is not None:
            return band_index
        elif len(self.array.shape) == 3:
            return tuple(x + 1 for x in range(len(self.array)))
        elif len(self.array.shape) == 2:
            return 1
        else:
            raise ValueError

    def _add_meta_from_src(self, src):
        if not hasattr(self, "_bounds"):
            self._bounds = tuple(src.bounds)
        self._width = src.width
        self._height = src.height

        if not hasattr(self, "_band_index") or self._band_index is None:
            self._band_index = src.indexes

        if not hasattr(self, "nodata") or self.nodata is None:
            self.nodata = src.nodata

        try:
            self._crs = pyproj.CRS(src.crs)
        except pyproj.exceptions.CRSError:
            self._crs = None

    def _load_warp_file(self) -> DatasetReader:
        """(from Torchgeo). Load and warp a file to the correct CRS and resolution.

        Args:
            filepath: file to load and warp

        Returns:
            file handle of warped VRT
        """
        with opener(self.path) as file:
            src = rasterio.open(file)

        # Only warp if necessary
        if src.crs != self.crs:
            vrt = WarpedVRT(src, crs=self.crs)
            src.close()
            return vrt
        return src

    def _read_tif(self, **kwargs) -> None:
        def _read(self, src):
            self._add_meta_from_src(src)
            out_shape = self.get_shape_from_res(kwargs.pop("res", None))

            if hasattr(self, "_warped_crs"):
                src = WarpedVRT(src, crs=self.crs)

            self.array = src.read(
                out_shape=out_shape,
                **self.read_kwargs(kwargs),
            )
            if self._dtype:
                self = self.astype(self._dtype)
            else:
                self = self.as_minimum_dtype()

        with opener(self.path) as file:
            with rasterio.open(file) as src:
                _read(self, src)

    def _read_with_mask(self, mask, boundless, **kwargs):
        kwargs = {"mask": mask, "boundless": boundless} | kwargs

        def _read(self, src, mask, boundless, **kwargs):
            # to non-warped crs
            # mask = mask.to_crs(self._crs)

            self._add_meta_from_src(src)
            transform = self.transform
            window = rasterio.windows.from_bounds(*to_bbox(mask), transform=transform)
            kwargs = {
                "window": window,
                "boundless": boundless,
                **self.read_kwargs(kwargs),
            }
            if hasattr(self, "_warped_crs"):
                src = WarpedVRT(src, crs=self.crs)
                kwargs.pop("boundless")
            self.array = src.read(**kwargs)

            self._bounds = src.window_bounds(window=window)

            if not np.size(self.array):
                return

            if self._dtype:
                self = self.astype(self._dtype)
            else:
                self = self.as_minimum_dtype()

        if self.array is not None:
            with memfile_from_array(self.array, **self.profile) as src:
                _read(self, src, **kwargs)
            return

        with opener(self.path) as file:
            with rasterio.open(file, **self.profile) as src:
                _read(self, src, **kwargs)

    def get_shape_from_res(self, res):
        if res is None:
            return None
        if hasattr(res, "__iter__") and len(res) == 2:
            res, res = res
        diffx = self.bounds[2] - self.bounds[0]
        diffy = self.bounds[3] - self.bounds[1]
        width = int(diffx / res)
        height = int(diffy / res)
        if hasattr(self.band_index, "__iter__"):
            return len(self.band_index), width, height
        return width, height

    def _write(self, dst, window):
        if np.ma.is_masked(self.array):
            if len(self.array.shape) == 2:
                return dst.write(
                    self.array.filled(self.nodata), indexes=1, window=window
                )

            for i in range(len(self.band_index_tuple())):
                dst.write(
                    self.array[i].filled(self.nodata),
                    indexes=i + 1,
                    window=window,
                )

        else:
            if len(self.array.shape) == 2:
                return dst.write(self.array, indexes=1, window=window)

            for i, idx in enumerate(self.band_index_tuple()):
                dst.write(self.array[i], indexes=idx, window=window)

    def _get_band_index(self, band_index):
        if isinstance(band_index, numbers.Number):
            return int(band_index)
        if band_index is None:
            return None
        try:
            return tuple(int(x) for x in band_index)
        except Exception as e:
            raise TypeError(
                "band_index should be an integer or an iterable of integers."
            ) from e

    def _return_gdf(self, obj) -> GeoDataFrame:
        if isinstance(obj, str) and not is_wkt(obj):
            return self._read_tif(obj)
        elif isinstance(obj, Raster):
            return obj.to_gdf()
        elif is_bbox_like(obj):
            return to_gdf(shapely.box(*to_bbox(obj)), crs=self.crs)
        else:
            return to_gdf(obj, crs=self.crs)

    @staticmethod
    def _to_geojson(gdf: GeoDataFrame) -> list[dict]:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            return [x["geometry"] for x in loads(gdf.to_json())["features"]]

    @staticmethod
    def _to_geojson_geom_val(gdf: GeoDataFrame, column: str) -> list[dict]:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            return [
                (feature["geometry"], val)
                for val, feature in zip(gdf[column], loads(gdf.to_json())["features"])
            ]

    @staticmethod
    def _array_to_geojson(array: np.ndarray, transform: Affine):
        try:
            return [
                (value, shape(geom))
                for geom, value in features.shapes(array, transform=transform)
            ]
        except ValueError:
            array = array.astype(np.float32)
            return [
                (value, shape(geom))
                for geom, value in features.shapes(array, transform=transform)
            ]

    def min(self):
        if np.size(self.array):
            return np.min(self.array)
        return None

    def max(self):
        if np.size(self.array):
            return np.max(self.array)
        return None

    def __hash__(self):
        return hash(self._hash)

    def __eq__(self, other):
        if isinstance(other, Raster):
            return np.all(np.array_equal(self.array, other.array))
        return NotImplemented

    def __repr__(self) -> str:
        """The print representation."""
        shape = self.shape
        shp = ", ".join([str(x) for x in shape])
        try:
            res = int(self.res[0])
        except TypeError:
            res = None
        crs = str(self.crs_to_string(self.crs))
        raster_id = self.raster_id
        path = self.path
        return f"{self.__class__.__name__}(shape=({shp}), res={res}, raster_id={raster_id}, crs={crs}, path={path})"

    def __setattr__(self, __name: str, __value) -> None:
        return super().__setattr__(__name, __value)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __mul__(self, scalar):
        self.check_for_array()
        self.array = self.array * scalar
        return self

    def __add__(self, scalar):
        self.check_for_array()
        self.array = self.array + scalar
        return self

    def __sub__(self, scalar):
        self.check_for_array()
        self.array = self.array - scalar
        return self

    def __truediv__(self, scalar):
        self.check_for_array()
        self.array = self.array / scalar
        return self

    def __floordiv__(self, scalar):
        self.check_for_array()
        self.array = self.array // scalar
        return self

    def __pow__(self, exponent):
        self.check_for_array()
        self.array = self.array**exponent
        return self

    def _return_self_or_copy(self, array, copy: bool):
        if not copy:
            self.array = array
            return self
        else:
            copy = self.copy()
            copy.array = array
            return copy



import geopandas as gpd
import numpy as np

from ..geopandas_tools.bounds import get_total_bounds
from ..helpers import unit_is_degrees
from .maps import clipmap, explore, samplemap


class Examine:
    """Explore geometries one row at a time.

    It takes one or more GeoDataFrames and shows an interactive map
    of one area at the time with the 'next' method or random areas with
    the 'sample' method.

    After creating the examiner object, the 'next' method will create a map
    showing all geometries within a given radius (the size parameter) of the
    first geometry in 'mask_gdf' (or the first speficied gdf). The 'next' method
    can then be repeated.

    Args:
        *gdfs: One or more GeoDataFrames. The rows of the first GeoDataFrame
            will be used as masks, unless 'mask_gdf' is specified.
        column: Column to use as colors.
        mask_gdf: Optional GeoDataFrame to use as mask iterator. The geometries
            of mask_gdf will not be shown.
        size: Number of meters (or other crs unit) to buffer the mask geometry
            before clipping.
        sort_values: Optional sorting column(s) of the mask GeoDataFrame. Rows
            will be iterated through from the top.
        **kwargs: Additional keyword arguments passed to sgis.clipmap.

    Examples
    --------
    Create the examiner.

    >>> import sgis as sg
    >>> roads = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> points = sg.read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> e = sg.Examine(points, roads)
    >>> e
    Examine(indices=1000, current=0, n_gdfs=2)

    Then the line below can be repeated for all rows if 'points'. This has to be
    in a separate notebook cell to the previous.

    >>> e.next()
    i == 1 (of 1000)
    <folium.folium.Map object at 0x000002AC73ACC090>

    Previous geometry:

    >>> e.next(-1)
    i == 0 (of 1000)
    <folium.folium.Map object at 0x0000020F3D68BE50>

    Repeating -1 will display the last row of 'points'.

    >>> e.next(-1)
    i == -1 (of 1000)
    <folium.folium.Map object at 0x0000020F3E46FB50>

    Show index 100 and color the map by 'idx':

    >>> e.next(100, column="idx")
    i == 100 (of 1000)
    <folium.folium.Map object at 0x0000020F3DD73F50>

    """

    def __init__(
        self,
        *gdfs: gpd.GeoDataFrame,
        column: str | None = None,
        mask_gdf: gpd.GeoDataFrame | None = None,
        sort_values: str | None = None,
        size: int | float = 1000,
        only_show_mask: bool = False,
        **kwargs,
    ):
        if not all(isinstance(gdf, gpd.GeoDataFrame) or not len(gdf) for gdf in gdfs):
            raise ValueError("gdfs must be of type GeoDataFrame.")

        self._gdfs = gdfs
        if mask_gdf is None:
            self.mask_gdf = gdfs[0]
        else:
            self.mask_gdf = mask_gdf

        self.indices = list(range(len(self.mask_gdf)))
        self.i = 0
        self.column = column
        self.size = size
        self.kwargs = kwargs

        if not len(self.mask_gdf):
            return

        if unit_is_degrees(self.mask_gdf) and size > 360:
            raise ValueError(
                "CRS unit is degrees. Use geopandas' "
                "to_crs method to change crs to e.g. UTM. "
                "Or set 'size' to a smaller number."
            )

        if sort_values is not None:
            if (
                sort_values == "area"
                or not isinstance(sort_values, str)
                and "area" in sort_values
            ):
                self.mask_gdf["area"] = self.mask_gdf.area
            self.mask_gdf = self.mask_gdf.sort_values(sort_values)

        if only_show_mask:
            self.kwargs["show"] = [True] + [False] * len(self._gdfs[1:])
        elif not kwargs.get("show"):
            self.kwargs["show"] = [True] * len(self._gdfs)

    def add_gdfs(self, *gdfs):
        self._gdfs = self._gdfs + gdfs
        self.kwargs["show"] += [True for _ in range(len(gdfs))]
        return self

    def next(self, i: int | None = None, **kwargs):
        """Displays a map of geometries within the next row of the mask gdf.

        Args:
            i: Index to display.
            **kwargs: Additional keyword arguments passed to sgis.clipmap.

        """
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        if i and i < 0:
            self.i += i - 1
        elif i:
            self.i = i

        if self.i >= len(self.mask_gdf):
            print("All rows are shown.")
            return

        print(f"i == {self.i} (of {len(self.mask_gdf)})")
        clipmap(
            *self._gdfs,
            self.column,
            mask=self.mask_gdf.iloc[[self.i]].buffer(self.size),
            **self.kwargs,
        )
        self.i += 1

    def sample(self, **kwargs):
        """Takes a sample index of the mask and displays a map of this area.

        Args:
            **kwargs: Additional keyword arguments passed to sgis.clipmap.
        """
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        i = np.random.randint(0, len(self.mask_gdf))

        print(f"Showing index {i}")
        clipmap(
            *self._gdfs,
            self.column,
            mask=self.mask_gdf.iloc[[i]].buffer(self.size),
            **self.kwargs,
        )

    def current(self, i: int | None = None, **kwargs):
        """Repeat the last shown map."""
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        if i and i < 0:
            self.i -= i
        elif i:
            self.i = i

        print(f"{self.i + 1} of {len(self.mask_gdf)}")
        clipmap(
            *self._gdfs,
            self.column,
            mask=self.mask_gdf.iloc[[self.i]].buffer(self.size),
            **self.kwargs,
        )

    def explore(self, **kwargs):
        """Show all rows like the function explore."""
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        explore(
            *self._gdfs,
            self.column,
            **self.kwargs,
        )

    def clipmap(self, **kwargs):
        """Show all rows like the function clipmap."""
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        clipmap(
            *self._gdfs,
            self.column,
            **self.kwargs,
        )

    def samplemap(self, **kwargs):
        """Show all rows like the function samplemap."""
        if kwargs:
            kwargs = self._fix_kwargs(kwargs)
            self.kwargs = self.kwargs | kwargs

        samplemap(
            *self._gdfs,
            self.column,
            **self.kwargs,
        )

    @property
    def mask(self) -> gpd.GeoDataFrame:
        """Returns a GeoDataFrame of the last shown mask geometry."""
        return self.mask_gdf.iloc[[self.i]]

    @property
    def gdfs(self) -> tuple[gpd.GeoDataFrame]:
        """Returns all GeoDataFrames in the area of the last shown mask geometry."""
        mask = self.mask_gdf.iloc[[self.i]]
        gdfs = ()
        for gdf in self._gdfs:
            gdfs = gdfs + (gdf.clip(mask.buffer(self.size)),)
        return gdfs

    @property
    def bounds(self):
        return get_total_bounds(*self.gdfs)

    def _fix_kwargs(self, kwargs) -> dict:
        self.size = kwargs.pop("size", self.size)
        self.column = kwargs.pop("column", self.column)
        return kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(indices={len(self.indices)}, current={self.i}, n_gdfs={len(self._gdfs)})"

    def __add__(self, scalar):
        self.i += scalar
        return self

    def __sub__(self, scalar):
        self.i -= scalar
        return self



from collections.abc import Iterable

import networkx as nx
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from shapely import STRtree, difference, make_valid, unary_union

from .general import _determine_geom_type_args, _push_geom_col, clean_geoms
from .geometry_types import get_geom_type, make_all_singlepart, to_single_geom_type
from .overlay import clean_overlay


def update_geometries(
    gdf: GeoDataFrame,
    geom_type: str | None = None,
    keep_geom_type: bool | None = None,
    grid_size: int | None = None,
) -> GeoDataFrame:
    """Puts geometries on top of each other rowwise.

    Since this operation is done rowwise, it's important to
    first sort the GeoDataFrame approriately. See example below.

    Args:
        gdf: The GeoDataFrame to be updated.
        keep_geom_type: If True, return only geometries of original type in case
            of intersection resulting in multiple geometry types or
            GeometryCollections. If False, return all resulting geometries
            (potentially mixed types).
        geom_type: Optionally specify what geometry type to keep.,
            if there are mixed geometry types. Must be either "polygon",
            "line" or "point".
        grid_size: Precision grid size to round the geometries. Will use the highest
            precision of the inputs by default.

    Example
    ------
    Create two circles and get the overlap.

    >>> import sgis as sg
    >>> circles = sg.to_gdf([(0, 0), (1, 1)]).pipe(sg.buff, 1)
    >>> duplicates = sg.get_intersections(circles)
    >>> duplicates
       idx                                           geometry
    0    1  POLYGON ((0.03141 0.99951, 0.06279 0.99803, 0....
    1    2  POLYGON ((1.00000 0.00000, 0.96859 0.00049, 0....

    The polygons are identical except for the order of the coordinates.

    >>> poly1, poly2 = duplicates.geometry
    >>> poly1.equals(poly2)
    True

    'update_geometries' gives different results based on the order
    of the GeoDataFrame.

    >>> sg.update_geometries(duplicates)
        idx                                           geometry
    0    1  POLYGON ((0.03141 0.99951, 0.06279 0.99803, 0....

    >>> dups_rev = duplicates.iloc[::-1]
    >>> sg.update_geometries(dups_rev)
        idx                                           geometry
    1    2  POLYGON ((1.00000 0.00000, 0.96859 0.00049, 0....

    It might be appropriate to put the largest polygons on top
    and sort all NaNs to the bottom.

    >>> updated = (
    ...     sg.sort_large_first(duplicates)
    ...     .pipe(sg.sort_nans_last)
    ...     .pipe(sg.update_geometries)
    >>> updated
        idx                                           geometry
    0    1  POLYGON ((0.03141 0.99951, 0.06279 0.99803, 0....

    """
    if len(gdf) <= 1:
        return gdf

    gdf = make_all_singlepart(clean_geoms(gdf))

    gdf, geom_type, keep_geom_type = _determine_geom_type_args(
        gdf, geom_type, keep_geom_type
    )

    geom_col = gdf._geometry_column_name
    index_mapper = {i: idx for i, idx in enumerate(gdf.index)}
    gdf = gdf.reset_index(drop=True)

    tree = STRtree(gdf.geometry.values)
    left, right = tree.query(gdf.geometry.values, predicate="intersects")
    indices = pd.Series(right, index=left).loc[lambda x: x.index > x.values]

    # select geometries from 'right', index from 'left', dissolve by 'left'
    erasers = (
        pd.Series(gdf.geometry.loc[indices.values].values, index=indices.index)
        .groupby(level=0)
        .agg(lambda x: make_valid(unary_union(x, grid_size=grid_size)))
    )

    # match up the aggregated erasers by index
    erased = make_valid(
        difference(
            gdf.geometry.loc[erasers.index],
            erasers,
            grid_size=grid_size,
        )
    )

    gdf.loc[erased.index, geom_col] = erased

    gdf = gdf.loc[~gdf.is_empty]

    gdf.index = gdf.index.map(index_mapper)

    if keep_geom_type:
        gdf = to_single_geom_type(gdf, geom_type)

    return gdf


def get_intersections(
    gdf: GeoDataFrame, geom_type: str | None = None, keep_geom_type: bool | None = None
) -> GeoDataFrame:
    """Find geometries that intersect in a GeoDataFrame.

    Does an intersection with itself and keeps only the geometries that appear
    more than once.

    Note that the returned GeoDataFrame in most cases contain two rows per
    intersection pair. It might also contain more than two overlapping polygons
    if there were multiple overlapping. These can be removed with
    update_geometries. See example below.

    Args:
        gdf: GeoDataFrame of polygons.
        geom_type: Optionally specify which geometry type to keep.
            Either "polygon", "line" or "point".
        keep_geom_type: Whether to keep the original geometry type.
            If mixed geometry types and keep_geom_type=True,
            an exception is raised.

    Returns:
        A GeoDataFrame of the overlapping polygons.

    Examples
    --------
    Create three partially overlapping polygons.

    >>> import sgis as sg
    >>> circles = sg.to_gdf([(0, 0), (1, 0), (2, 0)]).pipe(sg.buff, 1.2)
    >>> circles.area
    0    4.523149
    1    4.523149
    2    4.523149
    dtype: float64

    Get the duplicates.

    >>> duplicates = sg.get_intersections(circles)
    >>> duplicates["area"] = duplicates.area
    >>> duplicates
                                                geometry      area
    0  POLYGON ((1.19941 -0.03769, 1.19763 -0.07535, ...  2.194730
    0  POLYGON ((1.19941 -0.03769, 1.19763 -0.07535, ...  0.359846
    1  POLYGON ((0.48906 -1.08579, 0.45521 -1.06921, ...  2.194730
    1  POLYGON ((2.19941 -0.03769, 2.19763 -0.07535, ...  2.194730
    2  POLYGON ((0.98681 -0.64299, 0.96711 -0.61085, ...  0.359846
    2  POLYGON ((1.48906 -1.08579, 1.45521 -1.06921, ...  2.194730

    We get two rows for each intersection pair.

    To get no overlapping geometries without , we can put geometries
    on top of each other rowwise.

    >>> updated = sg.update_geometries(duplicates)
    >>> updated["area"] = updated.area
    >>> updated
           area                                           geometry
    0  2.194730  POLYGON ((1.19941 -0.03769, 1.19763 -0.07535, ...
    1  1.834884  POLYGON ((2.19763 -0.07535, 2.19467 -0.11293, ...

    It might be appropriate to sort the dataframe by columns.
    Or put large polygons first and NaN values last.

    >>> updated = (
    ...     sg.sort_large_first(duplicates)
    ...     .pipe(sg.sort_nans_last)
    ...     .pipe(sg.update_geometries)
    ... )
    >>> updated
          area                                           geometry
    0  2.19473  POLYGON ((1.19941 -0.03769, 1.19763 -0.07535, ...
    1  2.19473  POLYGON ((2.19763 -0.07535, 2.19467 -0.11293, ...

    """
    if isinstance(gdf, GeoSeries):
        gdf = GeoDataFrame({"geometry": gdf}, crs=gdf.crs)
        was_geoseries = True
    else:
        was_geoseries = False

    gdf, geom_type, keep_geom_type = _determine_geom_type_args(
        gdf, geom_type, keep_geom_type
    )

    idx_name = gdf.index.name
    gdf = gdf.assign(orig_idx=gdf.index).reset_index(drop=True)

    duplicated_geoms = _get_intersecting_geometries(
        gdf, geom_type, keep_geom_type
    ).pipe(clean_geoms)

    duplicated_geoms.index = duplicated_geoms["orig_idx"].values
    duplicated_geoms.index.name = idx_name

    if was_geoseries:
        return duplicated_geoms.geometry

    return duplicated_geoms.drop(columns="orig_idx")


def _get_intersecting_geometries(
    gdf: GeoDataFrame, geom_type, keep_geom_type
) -> GeoDataFrame:
    right = gdf[[gdf._geometry_column_name]]
    right["idx_right"] = right.index

    left = (
        gdf
        if not any("index_" in str(col) for col in gdf)
        else gdf.loc[:, lambda x: x.columns.difference({"index_right", "index_left"})]
    )
    left["idx_left"] = left.index

    def are_not_identical(df):
        return df["idx_left"] != df["idx_right"]

    if geom_type or get_geom_type(gdf) != "mixed":
        intersected = clean_overlay(
            left,
            right,
            how="intersection",
            geom_type=geom_type,
            keep_geom_type=keep_geom_type,
        ).loc[are_not_identical]
    else:
        if keep_geom_type:
            raise ValueError(
                "Cannot set keep_geom_type=True when the geom_type is mixed."
            )
        gdf = make_all_singlepart(gdf)
        intersected = []
        for geom_type in ["polygon", "line", "point"]:
            if not len(to_single_geom_type(gdf, geom_type)):
                continue
            intersected += [
                clean_overlay(left, right, how="intersection", geom_type=geom_type)
            ]
        intersected = pd.concat(intersected, ignore_index=True).loc[are_not_identical]

    # make sure it's correct by sjoining a point inside the polygons
    points_joined = intersected.representative_point().to_frame().sjoin(intersected)

    duplicated_points = points_joined.loc[points_joined.index.duplicated(keep=False)]

    return intersected.loc[intersected.index.isin(duplicated_points.index)].drop(
        columns=["idx_left", "idx_right"]
    )


def _drop_duplicate_geometries(gdf: GeoDataFrame, **kwargs) -> GeoDataFrame:
    """Drop geometries that are considered equal.

    Args:
        gdf: GeoDataFrame.
        **kwargs: Keyword arguments passed to pandas.DataFrame.drop_duplicates

    Returns:
        The GeoDataFrame with duplicate geometries dropped.
    """
    return (
        _get_duplicate_geometry_groups(gdf, group_col="dup_idx_")
        .drop_duplicates("dup_idx_", **kwargs)
        .drop(columns="dup_idx_")
    )


def _get_duplicate_geometry_groups(
    gdf: GeoDataFrame, group_col: str | Iterable[str] = "duplicate_index"
):
    idx_mapper = dict(enumerate(gdf.index))
    idx_name = gdf.index.name

    gdf = gdf.reset_index(drop=True)

    tree = STRtree(gdf.geometry.values)
    left, right = tree.query(gdf.geometry.values, predicate="within")

    edges = list(zip(left, right))

    graph = nx.Graph()
    graph.add_edges_from(edges)

    component_mapper = {
        j: i
        for i, component in enumerate(nx.connected_components(graph))
        for j in component
    }

    gdf[group_col] = component_mapper

    gdf = _push_geom_col(gdf)

    gdf.index = gdf.index.map(idx_mapper)
    gdf.index.name = idx_name

    return gdf



"""Attributes of the legend of the ThematicMap class.

The Legend class is best accessed through the 'legend' attribute of the ThematicMap
class.

"""
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.lines import Line2D
from pandas import Series

from ..geopandas_tools.bounds import points_in_bounds


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


class Legend:
    """Holds the general attributes of the legend in the ThematicMap class.

    This class holds attributes of the 'legend' attribute of the ThematicMap class.
    The fontsize, title_fontsize and markersize attributes are adjusted
    according to the size attribute of the ThematicMap.

    If a numeric column is used, additional attributes can be found in the
    ContinousLegend class.

    Attributes:
        title: Legend title. Defaults to the column name if used in the
            ThematicMap class.
        position: The legend's x and y position in the plot, specified as a tuple of
            x and y position between 0 and 1. E.g. position=(0.8, 0.2) for a position
            in the bottom right corner, (0.2, 0.8) for the upper left corner.
        fontsize: Text size of the legend labels. Defaults to the size of
            the ThematicMap class.
        title_fontsize: Text size of the legend title. Defaults to the
            size * 1.2 of the ThematicMap class.
        markersize: Size of the color circles in the legend. Defaults to the size of
            the ThematicMap class.
        framealpha: Transparency of the legend background.
        edgecolor: Color of the legend border. Defaults to #0f0f0f (almost black).
        kwargs: Stores additional keyword arguments taken by the matplotlib legend
            method. Specify this as e.g. m.legend.kwargs["labelcolor"] = "red", where
            'm' is the name of the ThematicMap instance. See here:
            https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html

    Examples
    --------
    Create ten random points with a numeric column from 0 to 9.

    >>> import sgis as sg
    >>> points = sg.random_points(10)
    >>> points["number"] = range(10)
    >>> points
                    geometry  number
    0  POINT (0.59780 0.50425)       0
    1  POINT (0.07019 0.26167)       1
    2  POINT (0.56475 0.15422)       2
    3  POINT (0.87293 0.60316)       3
    4  POINT (0.47373 0.20040)       4
    5  POINT (0.98661 0.15614)       5
    6  POINT (0.30951 0.77057)       6
    7  POINT (0.47802 0.52824)       7
    8  POINT (0.12215 0.96588)       8
    9  POINT (0.02938 0.93467)       9

    Creating the ThematicMap instance will also create the legend. Since we
    specify a numeric column, a ContinousLegend instance is created.

    >>> m = sg.ThematicMap(points, column="number")
    >>> m.legend
    <sgis.maps.legend.ContinousLegend object at 0x00000222206738D0>

    Changing the attributes that apply to both numeric and categorical columns.

    >>> m.legend.title = "Meters"
    >>> m.legend.title_fontsize = 11
    >>> m.legend.fontsize = 9
    >>> m.legend.markersize = 7.5
    >>> m.legend.position = (0.35, 0.28)
    >>> m.plot()

    Additional matplotlib keyword arguments can be specified as kwargs.

    >>> m.legend.kwargs["labelcolor"] = "red"

    Since we are using a numeric column, the legend is of type ContinousLegend.
    We can therefore also access the attributes that only apply to numeric columns.

    >>> m.label_sep = "to"
    >>> m.label_suffix = "num"
    >>> m.rounding = 2
    >>> m.plot()

    """

    def __init__(
        self,
        title: str | None = None,
        labels: list[str] | None = None,
        position: tuple[float] | None = None,
        markersize: int | None = None,
        fontsize: int | None = None,
        title_fontsize: int | None = None,
        framealpha: float = 1.0,
        edgecolor: str = "#0f0f0f",
        **kwargs,
    ):
        self.title = title

        if "size" in kwargs:
            size = kwargs.pop("size")
            self._get_legend_sizes(size, kwargs)
        else:
            self._title_fontsize = title_fontsize
            self._fontsize = fontsize
            self._markersize = markersize

        self.framealpha = framealpha
        self.edgecolor = edgecolor
        self.width = kwargs.pop("width", 0.1)
        self.height = kwargs.pop("height", 0.1)
        self.title_color = kwargs.pop("title_color", None)
        self.labelspacing = kwargs.pop("labelspacing", 0.8)

        self.labels = labels
        self._position = position
        self.kwargs = kwargs
        self._position_has_been_set = True if position else False

    def _get_legend_sizes(self, size, kwargs):
        """Adjust fontsize and markersize to size kwarg."""

        if "title_fontsize" in kwargs:
            self._title_fontsize = kwargs["title_fontsize"]
            self._title_fontsize_has_been_set = True
        else:
            self._title_fontsize = size * 1.2

        if "fontsize" in kwargs:
            self._fontsize = kwargs["fontsize"]
            self._fontsize_has_been_set = True
        else:
            self._fontsize = size

        if "markersize" in kwargs:
            self._markersize = kwargs["markersize"]
            self._markersize_has_been_set = True
        else:
            self._markersize = size

    def _prepare_categorical_legend(self, categories_colors: dict, nan_label: str):
        for attr in self.__dict__.keys():
            if attr in self.kwargs:
                self[attr] = self.kwargs.pop(attr)

        # swap column values with label values if labels is dict
        if self.labels and isinstance(self.labels, dict):
            categories_colors = {
                self.labels[cat]: color for cat, color in categories_colors.items()
            }
        # swap column values with label list and hope it's in the correct order
        elif self.labels:
            categories_colors = {
                label: color
                for label, color in zip(
                    self.labels, categories_colors.values(), strict=True
                )
            }

        self._patches, self._categories = [], []
        for category, color in categories_colors.items():
            if category == nan_label:
                self._categories.append(nan_label)
            else:
                self._categories.append(category)
            self._patches.append(
                Line2D(
                    [0],
                    [0],
                    linestyle="none",
                    marker="o",
                    alpha=self.kwargs.get("alpha", 1),
                    markersize=self._markersize,
                    markerfacecolor=color,
                    markeredgewidth=0,
                )
            )

    def _actually_add_legend(self, ax):
        legend = ax.legend(
            self._patches,
            self._categories,
            fontsize=self._fontsize,
            title=self.title,
            title_fontsize=self._title_fontsize,
            bbox_to_anchor=self._position + (self.width, self.height),
            fancybox=False,
            framealpha=self.framealpha,
            edgecolor=self.edgecolor,
            labelspacing=self.labelspacing,
            **self.kwargs,
        )

        if self.title_color:
            plt.setp(legend.get_title(), color=self.title_color)

        return ax

    def _get_best_legend_position(self, gdf, k: int):
        minx, miny, maxx, maxy = gdf.total_bounds
        diffx = maxx - minx
        diffy = maxy - miny

        points = points_in_bounds(gdf, 30)
        gdf = gdf.loc[:, ~gdf.columns.str.contains("index|level_")]
        joined = points.sjoin_nearest(gdf, distance_col="nearest")

        max_distance = max(joined.nearest)

        best_position = joined.loc[joined.nearest == max_distance].drop_duplicates(
            "geometry"
        )

        bestx, besty = (
            best_position.geometry.x.iloc[0],
            best_position.geometry.y.iloc[0],
        )

        bestx_01 = (bestx - minx) / (diffx)
        besty_01 = (besty - miny) / (diffy)

        bestx_01 = 0.1 if bestx_01 < 0.5 else 0.90
        besty_01 = 0.0375 * k if besty_01 < 0.5 else 1

        return bestx_01, besty_01

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, ValueError, IndexError, AttributeError):
            return default

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, new_value: bool):
        self._position = new_value
        self._position_has_been_set = True

    @property
    def title_fontsize(self):
        return self._title_fontsize

    @title_fontsize.setter
    def title_fontsize(self, new_value: bool):
        self._title_fontsize = new_value
        self._title_fontsize_has_been_set = True

    @property
    def fontsize(self):
        return self._fontsize

    @fontsize.setter
    def fontsize(self, new_value: bool):
        self._fontsize = new_value
        self._fontsize_has_been_set = True

    @property
    def markersize(self):
        return self._markersize

    @markersize.setter
    def markersize(self, new_value: bool):
        self._markersize = new_value
        self._markersize_has_been_set = True


class ContinousLegend(Legend):
    """Holds the legend attributes specific to numeric columns.

    The attributes consern the labeling of the groups in the legend.
    Labels can be set manually with the 'labels' attribute, or the format
    of the labels can be changed with the remaining attributes.

    Attributes:
        labels: To manually set labels. If set, all other labeling attributes are
            ignored. Should be given as a list of strings with the same length as
            the number of color groups.
        pretty_labels: If False (default), the minimum and maximum values of each
            color group will be used as legend labels. If True, the labels will end
            with the maximum value, but start at 1 + the maximum value of the previous
            group. The labels will be correct but inaccurate.
        label_suffix: The text to put after each number in the legend labels.
            Defaults to None.
        label_sep: Text to put in between the two numbers in each color group in
            the legend. Defaults to '-'.
        rounding: Number of decimals in the labels. By default, the rounding
            depends on the column's maximum value and standard deviation.
            OBS: The bins will not be rounded, meaning the labels might be wrong
            if not bins are set manually.
        thousand_sep: Separator between each thousand for large numbers. Defaults to
            None, meaning no separator.
        decimal_mark: Text to use as decimal point. Defaults to None, meaning '.' (dot)
            unless 'thousand_sep' is '.'. In this case, ',' (comma) will be used as
            decimal mark.

    Examples
    --------
    Create ten random points with a numeric column from 0 to 9.

    >>> import sgis as sg
    >>> points = sg.random_points(10)
    >>> points["number"] = range(10)
    >>> points
                    geometry  number
    0  POINT (0.59780 0.50425)       0
    1  POINT (0.07019 0.26167)       1
    2  POINT (0.56475 0.15422)       2
    3  POINT (0.87293 0.60316)       3
    4  POINT (0.47373 0.20040)       4
    5  POINT (0.98661 0.15614)       5
    6  POINT (0.30951 0.77057)       6
    7  POINT (0.47802 0.52824)       7
    8  POINT (0.12215 0.96588)       8
    9  POINT (0.02938 0.93467)       9

    Creating the ThematicMap instance with a numeric column.

    >>> m = sg.ThematicMap(points, column="number")

    Changing the attributes that apply to both numeric and categorical columns.

    >>> m.legend.title = "Meters"
    >>> m.legend.position = (0.35, 0.28)

    Change the attributes that only apply to numeric columns.

    >>> m.label_sep = "to"
    >>> m.label_suffix = "num"
    >>> m.rounding = 2
    >>> m.plot()

    Setting labels manually. For better control, it might be wise to also set the bins
    manually. The following bins will create a plot with the color groups
    0-2, 3-5, 6-7 and 8-9. The legend labels can then be set accordingly.

    >>> m = sg.ThematicMap(points, column="number")
    >>> m.bins = [0, 2, 5, 7, 9]
    >>> m.legend.labels = ["0 to 2 num", "3 to 5 num", "6 to 7 num", "8 to 9 num"]
    >>> m.plot()

    We will get the same groups if we exclude the first and last bin values. The
    minimum and maximum values will be filled anyway.

    >>> m = sg.ThematicMap(points, column="number")
    >>> m.bins = [2, 5, 7]
    >>> m.legend.labels = ["0 to 2 num", "3 to 5 num", "6 to 7 num", "8 to 9 num"]
    >>> m.plot()

    """

    def __init__(
        self,
        labels: list[str] | None = None,
        pretty_labels: bool = False,
        label_suffix: str | None = None,
        label_sep: str = "-",
        rounding: int | None = None,
        thousand_sep: str | None = None,
        decimal_mark: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.pretty_labels = pretty_labels
        self.thousand_sep = thousand_sep
        self.decimal_mark = decimal_mark

        self.label_sep = label_sep
        self.label_suffix = "" if not label_suffix else label_suffix
        self._rounding = rounding
        self._rounding_has_been_set = True if rounding else False

    def _get_rounding(self, array: Series | np.ndarray) -> int:
        def isinteger(x):
            return np.equal(np.mod(x, 1), 0)

        if np.all(isinteger(array)):
            return 0

        closest_to_zero_idx = np.argmin(np.abs(array))
        closest_to_zero = np.abs(array[closest_to_zero_idx])

        between_1_and_0 = 1 > closest_to_zero > 0
        if between_1_and_0:
            return int(abs(np.log10(abs(closest_to_zero)))) + 1

        std_ = np.std(array)
        max_ = np.max(array)
        if max_ > 30 and std_ > 5:
            return 0
        if max_ > 5 and std_ > 1:
            return 1
        if max_ > 1 and std_ > 0.1:
            return 2
        return int(abs(np.log10(std_))) + 1

    @staticmethod
    def _set_rounding(bins, rounding: int | float):
        if rounding == 0:
            return [int(round(bin, 0)) for bin in bins]
        else:
            return [round(bin, rounding) for bin in bins]

    def _remove_max_legend_value(self):
        if not self._legend:
            raise ValueError("Cannot modify legend before it is created.")

    def _prepare_continous_legend(
        self,
        bins: list[float],
        colors: list[str],
        nan_label: str,
        bin_values: dict,
    ):
        # TODO: clean up this messy method

        for attr in self.__dict__.keys():
            if attr in self.kwargs:
                self[attr] = self.kwargs.pop(attr)

        self._patches, self._categories = [], []

        for color in colors:
            self._patches.append(
                Line2D(
                    [0],
                    [0],
                    linestyle="none",
                    marker="o",
                    alpha=self.kwargs.get("alpha", 1),
                    markersize=self._markersize,
                    markerfacecolor=color,
                    markeredgewidth=0,
                )
            )

        if self.labels:
            if len(self.labels) != len(colors):
                raise ValueError(
                    "Label list must be same length as the number of groups. "
                    f"Got k={len(colors)} and labels={len(colors)}."
                    f"labels: {', '.join(self.labels)}"
                    f"colors: {', '.join(colors)}"
                )
            self._categories = self.labels

        elif len(bins) == len(colors):
            for i, _ in enumerate(bins):
                min_ = np.min(bin_values[i])
                max_ = np.max(bin_values[i])
                try:
                    min_rounded = self._set_rounding([min_], self._rounding)[0]
                    max_rounded = self._set_rounding([max_], self._rounding)[0]
                except ValueError as e:
                    if "nan" in str(e).lower():
                        min_rounded, max_rounded == "NaN", "NaN"
                if min_ == max_:
                    self._categories.append(f"{min_rounded} {self.label_suffix}")
                else:
                    self._categories.append(
                        f"{min_rounded} {self.label_suffix} {self.label_sep} "
                        f"{max_rounded} {self.label_suffix}"
                    )

        else:
            for i, (cat1, cat2) in enumerate(zip(bins[:-1], bins[1:], strict=True)):
                if nan_label in str(cat1) or nan_label in str(cat2):
                    self._categories.append(nan_label)
                    continue

                min_ = np.min(bin_values[i])
                max_ = np.max(bin_values[i])
                min_rounded = self._set_rounding([min_], self._rounding)[0]
                max_rounded = self._set_rounding([max_], self._rounding)[0]
                if self.pretty_labels:
                    if i != 0 and self._rounding == 0:
                        cat1 = int(cat1 + 1)
                    elif i != 0:
                        cat1 = cat1 + float(f"1e-{self._rounding}")

                    cat1 = self._format_number(cat1)
                    cat2 = self._format_number(cat2)

                    if min_ == max_:
                        label = self._two_value_label(cat1, cat2)
                        self._categories.append(label)
                        continue

                    label = self._two_value_label(cat1, cat2)
                    self._categories.append(label)

                elif min_ == max_:
                    min_rounded = self._format_number(min_rounded)
                    label = self._one_value_label(min_rounded)
                    self._categories.append(label)
                else:
                    min_rounded = self._format_number(min_rounded)
                    max_rounded = self._format_number(max_rounded)
                    label = self._two_value_label(min_rounded, max_rounded)
                    self._categories.append(label)

    def _actually_add_legend(self, ax):
        legend = ax.legend(
            self._patches,
            self._categories,
            fontsize=self._fontsize,
            title=self.title,
            title_fontsize=self._title_fontsize,
            bbox_to_anchor=self._position + (self.width, self.height),
            fancybox=False,
            framealpha=self.framealpha,
            edgecolor=self.edgecolor,
            labelspacing=self.labelspacing,
            **self.kwargs,
        )

        if self.title_color:
            plt.setp(legend.get_title(), color=self.title_color)

        return ax

    def _two_value_label(self, value1, value2):
        return (
            f"{value1} {self.label_suffix} {self.label_sep} "
            f"{value2} {self.label_suffix}"
        )

    def _one_value_label(self, value1):
        return f"{value1} {self.label_suffix}"

    def _format_number(self, number):
        if not self.thousand_sep and not self.decimal_mark:
            return number

        if self.thousand_sep:
            number = f"{number:,}".replace(",", "*temp_thousand*")

        if self.decimal_mark:
            number = str(number).replace(".", "*temp_decimal*")

        if self.thousand_sep == "." and not self.decimal_mark:
            number = number.replace(".", ",").replace(
                "*temp_thousand*", self.thousand_sep
            )
        elif not self.thousand_sep:
            number = number.replace("*temp_decimal*", self.decimal_mark)
        elif not self.decimal_mark:
            number = number.replace("*temp_thousand*", self.thousand_sep)
        else:
            number = number.replace("*temp_thousand*", self.thousand_sep).replace(
                "*temp_decimal*", self.decimal_mark
            )
        return number

    @property
    def rounding(self):
        return self._rounding

    @rounding.setter
    def rounding(self, new_value: bool):
        self._rounding = new_value
        self._rounding_has_been_set = True



import numpy as np
from geopandas import GeoDataFrame

from ..geopandas_tools.neighbors import get_k_nearest_neighbors
from .networkanalysisrules import NetworkAnalysisRules


"""
These are internal classes used in the NetworkAnalysis class. The classes used in
NetworkAnalysis are Origins and Destinations, which are subclasses of Points. The
Origins and Destinations classes are the same, except that the edges are in
opposite directions.
"""


class Points:
    def __init__(
        self,
        points: GeoDataFrame,
    ) -> None:
        self.gdf = points.copy()

    def _make_temp_idx(self, start: int) -> None:
        """Make a temporary id column thad don't overlap with the node ids.

        The original indices are stored in a dict and mapped back to the results in the
        end.
        """

        self.gdf["temp_idx"] = np.arange(start=start, stop=start + len(self.gdf))
        self.gdf["temp_idx"] = self.gdf["temp_idx"].astype(str)

        self.idx_dict = {
            temp_idx: idx
            for temp_idx, idx in zip(self.gdf.temp_idx, self.gdf.index, strict=True)
        }

    @staticmethod
    def _convert_distance_to_weight(distances, rules):
        """Meters to minutes based on 'weight_to_nodes_' attribute of the rules."""
        if not rules.nodedist_multiplier and not rules.nodedist_kmh:
            return [0 for _ in distances]

        if rules.nodedist_multiplier and rules.nodedist_kmh:
            raise ValueError(
                "Can only specify one of 'nodedist_multiplier' and 'nodedist_kmh'"
            )

        if rules.nodedist_multiplier:
            if rules.weight != "meters":
                raise ValueError(
                    "Can only specify 'nodedist_multiplier' when the 'weight' is meters"
                )
            return [x * rules.nodedist_multiplier for x in distances]

        if rules.nodedist_kmh and rules.weight != "minutes":
            raise ValueError(
                "Can only specify 'nodedist_kmh' when the 'weight' is minutes"
            )

        return [x / (16.666667 * rules.nodedist_kmh) for x in distances]

    def _make_edges(self, df, from_col, to_col):
        return [(f, t) for f, t in zip(df[from_col], df[to_col], strict=True)]

    def _get_edges_and_weights(
        self,
        nodes: GeoDataFrame,
        rules: NetworkAnalysisRules,
        k: int,
        from_col: str,
        to_col: str,
    ):
        distances = get_k_nearest_neighbors(
            gdf=self.gdf.set_index("temp_idx"),
            neighbors=nodes.set_index("node_id"),
            k=k,
        )

        distances["distance_min"] = distances.groupby(level=0)["distance"].min()

        distances = distances.reset_index()

        search_factor_multiplier = 1 + rules.search_factor / 100
        distances = distances.loc[
            lambda df: (df.distance <= rules.search_tolerance)
            & (
                df.distance
                <= df.distance_min * search_factor_multiplier + rules.search_factor
            )
        ]

        edges = self._make_edges(distances, from_col=from_col, to_col=to_col)

        weighs = self._convert_distance_to_weight(
            distances=list(distances.distance), rules=rules
        )

        return edges, weighs


class Origins(Points):
    def __init__(
        self,
        points: GeoDataFrame,
    ) -> None:
        super().__init__(points)

    def _get_edges_and_weights(
        self,
        nodes: GeoDataFrame,
        rules: NetworkAnalysisRules,
        k: int,
        from_col="temp_idx",
        to_col="neighbor_index",
    ):
        """Note: opposite order as in Destinations."""
        return super()._get_edges_and_weights(nodes, rules, k, from_col, to_col)


class Destinations(Points):
    def __init__(
        self,
        points: GeoDataFrame,
    ) -> None:
        super().__init__(points)

    def _get_edges_and_weights(
        self,
        nodes: GeoDataFrame,
        rules: NetworkAnalysisRules,
        k: int,
        from_col="neighbor_index",
        to_col="temp_idx",
    ):
        """Note: opposite order as in Origins."""
        return super()._get_edges_and_weights(nodes, rules, k, from_col, to_col)



"""Functions for polygon geometries."""

import networkx as nx
import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame, GeoSeries
from shapely import (
    STRtree,
    area,
    box,
    buffer,
    difference,
    get_exterior_ring,
    get_interior_ring,
    get_num_interior_rings,
    get_parts,
    is_empty,
    make_valid,
    polygons,
    unary_union,
)
from shapely.errors import GEOSException

from .duplicates import get_intersections
from .general import _push_geom_col, clean_geoms, get_grouped_centroids, to_lines
from .geometry_types import get_geom_type, make_all_singlepart, to_single_geom_type
from .neighbors import get_neighbor_indices
from .overlay import clean_overlay
from .polygons_as_rings import PolygonsAsRings
from .sfilter import sfilter, sfilter_inverse


def get_polygon_clusters(
    *gdfs: GeoDataFrame | GeoSeries,
    cluster_col: str = "cluster",
    allow_multipart: bool = False,
    predicate: str | None = "intersects",
    as_string: bool = False,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    """Find which polygons overlap without dissolving.

    Devides polygons into clusters in a fast and precice manner by using spatial join
    and networkx to find the connected components, i.e. overlapping geometries.
    If multiple GeoDataFrames are given, the clusters will be based on all
    combined.

    This can be used instead of dissolve+explode, or before dissolving by the cluster
    column. This has been tested to be a lot faster if there are many
    non-overlapping polygons, but somewhat slower than dissolve+explode if most
    polygons overlap.

    Args:
        gdfs: One or more GeoDataFrames of polygons.
        cluster_col: Name of the resulting cluster column.
        allow_multipart: Whether to allow mutipart geometries in the gdfs.
            Defaults to False to avoid confusing results.
        as_string: Whether to return the cluster column values as a string with x and y
            coordinates. Convinient to always get unique ids.
            Defaults to False because of speed.

    Returns:
        One or more GeoDataFrames (same amount as was given) with a new cluster column.

    Examples
    --------

    Create geometries with three clusters of overlapping polygons.

    >>> import sgis as sg
    >>> gdf = sg.to_gdf([(0, 0), (1, 1), (0, 1), (4, 4), (4, 3), (7, 7)])
    >>> buffered = sg.buff(gdf, 1)
    >>> gdf
                                                geometry
    0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    Add a cluster column to the GeoDataFrame:

    >>> gdf = sg.get_polygon_clusters(gdf, cluster_col="cluster")
    >>> gdf
       cluster                                           geometry
    0        0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1        0  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2        0  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3        1  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4        1  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    If multiple GeoDataFrames are given, all are returned with common
    cluster values.

    >>> gdf2 = sg.to_gdf([(0, 0), (7, 7)])
    >>> gdf, gdf2 = sg.get_polygon_clusters(gdf, gdf2, cluster_col="cluster")
    >>> gdf2
       cluster                 geometry
    0        0  POINT (0.00000 0.00000)
    1        2  POINT (7.00000 7.00000)
    >>> gdf
       cluster                                           geometry
    0        0  POLYGON ((1.00000 0.00000, 0.99951 -0.03141, 0...
    1        0  POLYGON ((2.00000 1.00000, 1.99951 0.96859, 1....
    2        0  POLYGON ((1.00000 1.00000, 0.99951 0.96859, 0....
    3        1  POLYGON ((5.00000 4.00000, 4.99951 3.96859, 4....
    4        1  POLYGON ((5.00000 3.00000, 4.99951 2.96859, 4....
    5        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....

    Dissolving 'by' the cluster column will make the dissolve much
    faster if there are a lot of non-overlapping polygons.

    >>> dissolved = gdf.dissolve(by="cluster", as_index=False)
    >>> dissolved
       cluster                                           geometry
    0        0  POLYGON ((0.99951 -0.03141, 0.99803 -0.06279, ...
    1        1  POLYGON ((4.99951 2.96859, 4.99803 2.93721, 4....
    2        2  POLYGON ((8.00000 7.00000, 7.99951 6.96859, 7....
    """
    if isinstance(gdfs[-1], str):
        *gdfs, cluster_col = gdfs

    concated = []
    orig_indices = ()

    # take a copy only if there are gdfs with the same id
    # To not get any overwriting in the for loop
    if sum(df1 is df2 for df1 in gdfs for df2 in gdfs) > len(gdfs):
        gdfs = [gdf.copy() for gdf in gdfs]

    for i, gdf in enumerate(gdfs):
        if isinstance(gdf, GeoSeries):
            gdf = gdf.to_frame()

        if not isinstance(gdf, GeoDataFrame):
            raise TypeError("'gdfs' should be GeoDataFrames or GeoSeries.")

        if not allow_multipart and len(gdf) != len(gdf.explode(index_parts=False)):
            raise ValueError(
                "All geometries should be exploded to singlepart "
                "in order to get correct polygon clusters. "
                "To allow multipart geometries, set allow_multipart=True"
            )

        orig_indices = orig_indices + (gdf.index,)

        gdf = gdf.assign(i__=i)

        concated.append(gdf)

    concated = pd.concat(concated, ignore_index=True)

    if not len(concated):
        return concated.drop("i__", axis=1).assign(**{cluster_col: []})

    concated[cluster_col] = get_cluster_mapper(concated, predicate)

    if as_string:
        concated[cluster_col] = get_grouped_centroids(concated, groupby=cluster_col)

    concated = _push_geom_col(concated)

    n_gdfs = concated["i__"].unique()

    if len(n_gdfs) == 1:
        concated.index = orig_indices[0]
        return concated.drop(["i__"], axis=1)

    unconcated = ()
    for i in n_gdfs:
        gdf = concated[concated["i__"] == i]
        gdf.index = orig_indices[i]
        gdf = gdf.drop(["i__"], axis=1)
        unconcated = unconcated + (gdf,)

    return unconcated


def get_cluster_mapper(gdf, predicate="intersects"):
    if not gdf.index.is_unique:
        raise ValueError("Index must be unique")
    neighbors = get_neighbor_indices(gdf, gdf, predicate=predicate)

    edges = [(source, target) for source, target in neighbors.items()]

    graph = nx.Graph()
    graph.add_edges_from(edges)

    return {
        j: i
        for i, component in enumerate(nx.connected_components(graph))
        for j in component
    }


def eliminate_by_longest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    remove_isolated: bool = False,
    fix_double: bool = True,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    grid_size=None,
    **kwargs,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    """Dissolves selected polygons with the longest bordering neighbor polygon.

    Eliminates selected geometries by dissolving them with the neighboring
    polygon with the longest shared border. The index and column values of the
    large polygons will be kept, unless else is specified.

    Note that this might be a lot slower than eliminate_by_largest.

    Args:
        gdf: GeoDataFrame with polygon geometries, or a list of GeoDataFrames.
        to_eliminate: The geometries to be eliminated by 'gdf'.
        remove_isolated: If False (default), polygons in 'to_eliminate' that share
            no border with any polygon in 'gdf' will be kept. If True, the isolated
            polygons will be removed.
        fix_double: If True, geometries to be eliminated will be erased by overlapping
            geometries to not get double surfaces if the geometries in 'to_eliminate'
            overlaps with multiple geometries in 'gdf'.
        ignore_index: If False (default), the resulting GeoDataFrame will keep the
            index of the large polygons. If True, the resulting axis will be labeled
            0, 1, …, n - 1.
        aggfunc: Aggregation function(s) to use when dissolving/eliminating.
            Defaults to None, meaning the values of 'gdf' is used. Otherwise,
            aggfunc will be passed to pandas groupby.agg. note: The geometries of
            'gdf' are sorted first, but if 'gdf' has missing values, the resulting
            polygons might get values from the polygons to be eliminated
            (if aggfunc="first").
        kwargs: Keyword arguments passed to the dissolve method.

    Returns:
        The GeoDataFrame (gdf) with the geometries of 'to_eliminate' dissolved in.
        If multiple GeoDataFrame are passed as 'gdf', they are returned as a tuple.

    Examples
    --------

    Create two polygons with a sliver in between:

    >>> sliver = sg.to_gdf(Polygon([(0, 0), (0.1, 1), (0, 2), (-0.1, 1)]))
    >>> small_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (-0.1, 1), (0, 2), (-1, 2), (-2, 2), (-1, 1)])
    ... )
    >>> large_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (0.1, 1), (1, 2), (2, 2), (3, 2), (3, 0)])
    ... )

    Using multiple GeoDataFrame as input, the sliver is eliminated into the small
    polygon (because it has the longest border with sliver).

    >>> small_poly_eliminated, large_poly_eliminated = sg.eliminate_by_longest(
    ...     [small_poly, large_poly], sliver
    ... )

    With only one input GeoDataFrame:

    >>> polys = pd.concat([small_poly, large_poly])
    >>> eliminated = sg.eliminate_by_longest(polys, sliver)


    """
    if isinstance(gdf, (list, tuple)):
        # concat, then break up the dataframes in the end
        was_multiple_gdfs = True
        original_cols = [df.columns for df in gdf]
        gdf = pd.concat(df.assign(**{"_df_idx": i}) for i, df in enumerate(gdf))
    else:
        was_multiple_gdfs = False

    crs = gdf.crs
    geom_type = get_geom_type(gdf)

    if not ignore_index:
        idx_mapper = dict(enumerate(gdf.index))
        idx_name = gdf.index.name

    gdf = gdf.reset_index(drop=True)

    gdf["_dissolve_idx"] = gdf.index
    to_eliminate = to_eliminate.assign(_eliminate_idx=lambda x: range(len(x)))

    # convert to lines to get the borders
    lines_eliminate = to_lines(to_eliminate[["_eliminate_idx", "geometry"]])

    borders = (
        gdf[["_dissolve_idx", "geometry"]]
        .overlay(lines_eliminate, keep_geom_type=False)
        .loc[lambda x: x["_eliminate_idx"].notna()]
    )

    borders["_length"] = borders.length

    # as DataFrame because GeoDataFrame constructor is expensive
    borders = pd.DataFrame(borders)

    longest_border = borders.sort_values("_length", ascending=False).drop_duplicates(
        "_eliminate_idx"
    )

    to_dissolve_idx = longest_border.set_index("_eliminate_idx")["_dissolve_idx"]
    to_eliminate["_dissolve_idx"] = to_eliminate["_eliminate_idx"].map(to_dissolve_idx)

    actually_eliminate = to_eliminate.loc[to_eliminate["_dissolve_idx"].notna()]

    isolated = to_eliminate.loc[to_eliminate["_dissolve_idx"].isna()]
    containing_eliminators = (
        pd.DataFrame(
            isolated.drop(columns="_dissolve_idx").sjoin(
                gdf[["_dissolve_idx", "geometry"]], predicate="contains"
            )
        )
        .drop(columns="index_right")
        .drop_duplicates("_eliminate_idx")
    )

    eliminated = _eliminate(
        pd.DataFrame(gdf),
        pd.concat([actually_eliminate, containing_eliminators]),
        aggfunc,
        crs,
        fix_double,
        grid_size=grid_size,
        **kwargs,
    )

    if not ignore_index:
        eliminated.index = eliminated.index.map(idx_mapper)
        eliminated.index.name = idx_name

    if not remove_isolated and len(isolated):
        eliminated = pd.concat([eliminated, isolated])

    eliminated = eliminated.drop(
        ["_dissolve_idx", "_length", "_eliminate_idx", "_dissolve_idx"],
        axis=1,
        errors="ignore",
    )

    out = GeoDataFrame(eliminated, geometry="geometry", crs=crs).pipe(clean_geoms)

    if geom_type != "mixed":
        out = to_single_geom_type(out, geom_type)

    out = out.reset_index(drop=True) if ignore_index else out

    if not was_multiple_gdfs:
        return out

    gdfs = ()
    for i, cols in enumerate(original_cols):
        df = out.loc[out["_df_idx"] == i, cols]
        gdfs += (df,)

    return gdfs


def eliminate_by_largest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    max_distance: int | float | None = None,
    remove_isolated: bool = False,
    fix_double: bool = False,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate: str = "intersects",
    grid_size=None,
    **kwargs,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    """Dissolves selected polygons with the largest neighbor polygon.

    Eliminates selected geometries by dissolving them with the neighboring
    polygon with the largest area. The index and column values of the
    large polygons will be kept, unless else is specified.

    Args:
        gdf: GeoDataFrame with polygon geometries, or a list of GeoDataFrames.
        to_eliminate: The geometries to be eliminated by 'gdf'.
        remove_isolated: If False (default), polygons in 'to_eliminate' that share
            no border with any polygon in 'gdf' will be kept. If True, the isolated
            polygons will be removed.
        fix_double: If True, geometries to be eliminated will be erased by overlapping
            geometries to not get double surfaces if the geometries in 'to_eliminate'
            overlaps with multiple geometries in 'gdf'.
        ignore_index: If False (default), the resulting GeoDataFrame will keep the
            index of the large polygons. If True, the resulting axis will be labeled
            0, 1, …, n - 1.
        aggfunc: Aggregation function(s) to use when dissolving/eliminating.
            Defaults to None, meaning the values of 'gdf' is used. Otherwise,
            aggfunc will be passed to pandas groupby.agg. note: The geometries of
            'gdf' are sorted first, but if 'gdf' has missing values, the resulting
            polygons might get values from the polygons to be eliminated
            (if aggfunc="first").
        predicate: Binary predicate passed to sjoin. Defaults to "intersects".
        kwargs: Keyword arguments passed to the dissolve method.

    Returns:
        The GeoDataFrame (gdf) with the geometries of 'to_eliminate' dissolved in.
        If multiple GeoDataFrame are passed as 'gdf', they are returned as a tuple.

    Examples
    --------

    Create two polygons with a sliver in between:

    >>> sliver = sg.to_gdf(Polygon([(0, 0), (0.1, 1), (0, 2), (-0.1, 1)]))
    >>> small_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (-0.1, 1), (0, 2), (-1, 2), (-2, 2), (-1, 1)])
    ... )
    >>> large_poly = sg.to_gdf(
    ...     Polygon([(0, 0), (0.1, 1), (1, 2), (2, 2), (3, 2), (3, 0)])
    ... )

    Using multiple GeoDataFrame as input, the sliver is eliminated into
    the large polygon.

    >>> small_poly_eliminated, large_poly_eliminated = sg.eliminate_by_largest(
    ...     [small_poly, large_poly], sliver
    ... )

    With only one input GeoDataFrame:

    >>> polys = pd.concat([small_poly, large_poly])
    >>> eliminated = sg.eliminate_by_largest(polys, sliver)

    """
    return _eliminate_by_area(
        gdf,
        to_eliminate=to_eliminate,
        remove_isolated=remove_isolated,
        max_distance=max_distance,
        ignore_index=ignore_index,
        sort_ascending=False,
        aggfunc=aggfunc,
        predicate=predicate,
        fix_double=fix_double,
        grid_size=grid_size,
        **kwargs,
    )


def eliminate_by_smallest(
    gdf: GeoDataFrame | list[GeoDataFrame],
    to_eliminate: GeoDataFrame,
    *,
    max_distance: int | float | None = None,
    remove_isolated: bool = False,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate: str = "intersects",
    fix_double: bool = False,
    grid_size=None,
    **kwargs,
) -> GeoDataFrame | tuple[GeoDataFrame]:
    return _eliminate_by_area(
        gdf,
        to_eliminate=to_eliminate,
        remove_isolated=remove_isolated,
        max_distance=max_distance,
        ignore_index=ignore_index,
        sort_ascending=True,
        aggfunc=aggfunc,
        predicate=predicate,
        fix_double=fix_double,
        grid_size=grid_size,
        **kwargs,
    )


def _eliminate_by_area(
    gdf: GeoDataFrame,
    to_eliminate: GeoDataFrame,
    remove_isolated: bool,
    max_distance: int | float | None,
    sort_ascending: bool,
    ignore_index: bool = False,
    aggfunc: str | dict | list | None = None,
    predicate="intersects",
    fix_double: bool = False,
    grid_size=None,
    **kwargs,
) -> GeoDataFrame:
    if isinstance(gdf, (list, tuple)):
        was_multiple_gdfs = True
        original_cols = [df.columns for df in gdf]
        gdf = pd.concat(df.assign(**{"_df_idx": i}) for i, df in enumerate(gdf))
    else:
        was_multiple_gdfs = False

    crs = gdf.crs
    geom_type = get_geom_type(gdf)

    if not ignore_index:
        idx_mapper = dict(enumerate(gdf.index))
        idx_name = gdf.index.name

    gdf = make_all_singlepart(gdf).reset_index(drop=True)
    to_eliminate = make_all_singlepart(to_eliminate).reset_index(drop=True)

    gdf["_area"] = gdf.area
    gdf["_dissolve_idx"] = gdf.index

    if max_distance:
        to_join = gdf[["_area", "_dissolve_idx", "geometry"]]
        to_join.geometry = to_join.buffer(max_distance)
        joined = to_eliminate.sjoin(to_join, predicate=predicate, how="left")
    else:
        joined = to_eliminate.sjoin(
            gdf[["_area", "_dissolve_idx", "geometry"]], predicate=predicate, how="left"
        )

    # as DataFrames because GeoDataFrame constructor is expensive
    joined = (
        pd.DataFrame(joined)
        .drop(columns="index_right")
        .sort_values("_area", ascending=sort_ascending)
        .loc[lambda x: ~x.index.duplicated(keep="first")]
    )

    gdf = pd.DataFrame(gdf)

    notna = joined.loc[lambda x: x["_dissolve_idx"].notna()]

    eliminated = _eliminate(
        gdf, notna, aggfunc, crs, fix_double=fix_double, grid_size=grid_size, **kwargs
    )

    if not ignore_index:
        eliminated.index = eliminated.index.map(idx_mapper)
        eliminated.index.name = idx_name

    if not remove_isolated:
        isolated = joined.loc[joined["_dissolve_idx"].isna()]
        if len(isolated):
            eliminated = pd.concat([eliminated, isolated])

    eliminated = eliminated.drop(
        ["_dissolve_idx", "_area", "_eliminate_idx", "_dissolve_idx"],
        axis=1,
        errors="ignore",
    )

    out = GeoDataFrame(eliminated, geometry="geometry", crs=crs).pipe(clean_geoms)

    if geom_type != "mixed":
        out = to_single_geom_type(out, geom_type)

    out = out.reset_index(drop=True) if ignore_index else out

    if not was_multiple_gdfs:
        return out

    gdfs = ()
    for i, cols in enumerate(original_cols):
        df = out.loc[out["_df_idx"] == i, cols]
        gdfs += (df,)

    return gdfs


def _eliminate(gdf, to_eliminate, aggfunc, crs, fix_double, grid_size, **kwargs):
    if not len(to_eliminate):
        return gdf

    in_to_eliminate = gdf["_dissolve_idx"].isin(to_eliminate["_dissolve_idx"])
    to_dissolve = gdf.loc[in_to_eliminate]
    not_to_dissolve = gdf.loc[~in_to_eliminate].set_index("_dissolve_idx")

    to_eliminate["_to_eliminate"] = 1

    if aggfunc is None:
        concatted = pd.concat(
            [to_dissolve, to_eliminate[["_to_eliminate", "_dissolve_idx", "geometry"]]]
        )
        aggfunc = "first"
    else:
        concatted = pd.concat([to_dissolve, to_eliminate])

    one_hit = concatted.loc[
        lambda x: (x.groupby("_dissolve_idx").transform("size") == 1)
        & (x["_dissolve_idx"].notna())
    ].set_index("_dissolve_idx")

    assert len(one_hit) == 0

    many_hits = concatted.loc[
        lambda x: x.groupby("_dissolve_idx").transform("size") > 1
    ]

    if not len(many_hits):
        return one_hit

    # aggregate all columns except geometry
    kwargs.pop("as_index", None)
    eliminated = (
        many_hits.drop(columns="geometry")
        .groupby("_dissolve_idx", **kwargs)
        .agg(aggfunc)
        .drop(["_area"], axis=1, errors="ignore")
    )

    # aggregate geometry
    if fix_double:
        assert eliminated.index.is_unique

        many_hits = many_hits.set_index("_dissolve_idx")
        many_hits["_row_idx"] = range(len(many_hits))

        # TODO kan dette fikses trygt med .duplicated og ~x.duplicated?
        eliminators: pd.Series = many_hits.loc[
            many_hits["_to_eliminate"] != 1, "geometry"
        ]
        to_be_eliminated = many_hits.loc[many_hits["_to_eliminate"] == 1]

        if 0:
            tree = STRtree(eliminators.values)
            left, right = tree.query(
                to_be_eliminated.geometry.values, predicate="intersects"
            )
            pairs = pd.Series(right, index=left).to_frame("right")
            pairs["_dissolve_idx"] = pairs.index.map(
                dict(enumerate(to_be_eliminated.index))
            )

            soon_erased = to_be_eliminated.iloc[pairs.index]
            intersecting = eliminators.iloc[pairs["right"]]

            intersecting.index = soon_erased.index
            soon_erased = soon_erased.geometry.groupby(level=0).agg(unary_union)
            intersecting = intersecting.groupby(level=0).agg(unary_union)

            soon_erased.loc[:] = difference(
                soon_erased.values,
                intersecting.values,
            )
            intersecting.loc[:] = difference(
                intersecting.values,
                soon_erased.values,
            )

            eliminated["geometry"] = (
                pd.concat([intersecting, soon_erased])
                .groupby(level=0)
                .agg(lambda x: make_valid(unary_union(x.dropna().values)))
            )
            from ..maps.maps import explore, explore_locals

            explore_locals()

        # all_geoms: pd.Series = gdf.set_index("_dissolve_idx").geometry
        all_geoms: pd.Series = gdf.geometry

        tree = STRtree(all_geoms.values)
        left, right = tree.query(
            to_be_eliminated.geometry.values, predicate="intersects"
        )
        pairs = pd.Series(right, index=left).to_frame("right")
        pairs["_dissolve_idx"] = pairs.index.map(
            dict(enumerate(to_be_eliminated.index))
        )

        # pairs = pairs.loc[lambda x: x["right"] != x["_dissolve_idx"]]

        soon_erased = to_be_eliminated.iloc[pairs.index]
        intersecting = all_geoms.iloc[pairs["right"]]

        shoud_not_erase = soon_erased.index != intersecting.index
        soon_erased = soon_erased[shoud_not_erase]
        intersecting = intersecting[shoud_not_erase]

        missing = to_be_eliminated.loc[
            # (~to_be_eliminated.index.isin(soon_erased.index))
            # |
            (~to_be_eliminated["_row_idx"].isin(soon_erased["_row_idx"])),
            # | (~to_be_eliminated["_row_idx"].isin(soon_erased.index)),
            "geometry",
        ]

        if 0:
            from ..geopandas_tools.conversion import to_gdf
            from ..maps.maps import explore, explore_locals

            display(pairs)
            display(soon_erased.index.unique())
            display(soon_erased._row_idx.unique())
            display(to_be_eliminated.index.unique())
            display(to_be_eliminated._row_idx.unique())
            display(missing.index.unique())

            display(soon_erased)
            display(to_be_eliminated)
            display(missing)

            explore(
                to_gdf(soon_erased, 25833), intersecting=to_gdf(intersecting, 25833)
            )
            for j, ((i, g), (i2, g2)) in enumerate(
                zip(intersecting.items(), soon_erased.geometry.items())
            ):
                explore(
                    to_gdf(g, 25833).assign(ii=i, j=j),
                    g2=to_gdf(g2, 25833).assign(ii=i2, j=j),
                )

        if 0:
            explore(to_gdf(to_be_eliminated.iloc[[16]]))
            explore(to_gdf(to_be_eliminated.iloc[[15]]))
            explore(to_gdf(to_be_eliminated.iloc[[0]]))
            print("hei")
            explore(to_gdf(soon_erased.loc[soon_erased.index == 16]))
            explore(to_gdf(soon_erased.loc[soon_erased.index == 36]))

            explore(to_gdf(soon_erased.loc[soon_erased._row_idx == 16]))
            explore(to_gdf(soon_erased.loc[soon_erased._row_idx == 36]))

        # allign and aggregate by dissolve index to not get duplicates in difference
        intersecting.index = soon_erased.index
        soon_erased = soon_erased.geometry.groupby(level=0).agg(
            lambda x: unary_union(x, grid_size=grid_size)
        )
        intersecting = intersecting.groupby(level=0).agg(
            lambda x: unary_union(x, grid_size=grid_size)
        )

        # from ..maps.maps import explore_locals
        # explore_locals()

        soon_erased.loc[:] = difference(
            soon_erased.values,
            intersecting.values,
        )

        eliminated["geometry"] = (
            pd.concat([eliminators, soon_erased, missing])
            .groupby(level=0)
            .agg(
                lambda x: make_valid(
                    unary_union(x.dropna().values, grid_size=grid_size)
                )
            )
        )

    else:
        eliminated["geometry"] = many_hits.groupby("_dissolve_idx")["geometry"].agg(
            lambda x: make_valid(unary_union(x.values, grid_size=grid_size))
        )

    # setting crs on the GeometryArrays to avoid warning in concat
    not_to_dissolve.geometry.values.crs = crs
    eliminated.geometry.values.crs = crs
    one_hit.geometry.values.crs = crs

    to_concat = [not_to_dissolve, eliminated, one_hit]

    assert all(df.index.name == "_dissolve_idx" for df in to_concat)

    return pd.concat(to_concat).sort_index().drop(columns="_to_eliminate")


def close_thin_holes(gdf: GeoDataFrame, tolerance: int | float) -> GeoDataFrame:
    gdf = make_all_singlepart(gdf)
    holes = get_holes(gdf)
    inside_holes = sfilter(gdf, holes, predicate="within").unary_union

    def to_none_if_thin(geoms):
        try:
            buffered_in = buffer(
                difference(polygons(geoms), inside_holes), -(tolerance / 2)
            )
            return np.where(is_empty(buffered_in), None, geoms)
        except ValueError as e:
            if not len(geoms):
                return geoms
            raise e

    if not (gdf.geom_type == "Polygon").all():
        raise ValueError(gdf.geom_type.value_counts())

    return PolygonsAsRings(gdf).apply_numpy_func_to_interiors(to_none_if_thin).to_gdf()


def return_correct_geometry_object(in_obj, out_obj):
    if isinstance(in_obj, GeoDataFrame):
        in_obj.geometry = out_obj
        return in_obj
    elif isinstance(in_obj, GeoSeries):
        return GeoSeries(out_obj, crs=in_obj.crs)
    else:
        return out_obj


def close_all_holes(
    gdf: GeoDataFrame | GeoSeries,
    *,
    ignore_islands: bool = False,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries:
    """Closes all holes in polygons.

    It takes a GeoDataFrame or GeoSeries of polygons and
    returns the outer circle.

    Args:
        gdf: GeoDataFrame or GeoSeries of polygons.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.
        ignore_islands: If False (default), polygons inside the holes (islands)
            will be erased from the output geometries. If True, the entire
            holes will be closed and the islands kept, meaning there might be
            duplicate surfaces in the resulting geometries.
            Note that ignoring islands is a lot faster.

    Returns:
        A GeoDataFrame or GeoSeries of polygons with closed holes in the geometry
        column.

    Examples
    --------
    Let's create a circle with a hole in it.

    >>> point = sg.to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = sg.buff(point, 1000)
    >>> small_circle = sg.buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close the hole.

    >>> holes_closed = sg.close_all_holes(circle_with_hole)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise ValueError(
            f"'gdf' should be of type GeoDataFrame or GeoSeries. Got {type(gdf)}"
        )

    if not len(gdf):
        return gdf

    if copy:
        gdf = gdf.copy()

    gdf = make_all_singlepart(gdf)

    if ignore_islands:
        geoms = gdf.geometry if isinstance(gdf, GeoDataFrame) else gdf
        holes_closed = make_valid(polygons(get_exterior_ring(geoms)))
        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = holes_closed
            return gdf
        elif isinstance(gdf, GeoSeries):
            return GeoSeries(holes_closed, crs=gdf.crs)
        else:
            return holes_closed

    all_geoms = make_valid(gdf.unary_union)
    if isinstance(gdf, GeoDataFrame):
        gdf.geometry = gdf.geometry.map(
            lambda x: _close_all_holes_no_islands(x, all_geoms)
        )
        return gdf
    else:
        return gdf.map(lambda x: _close_all_holes_no_islands(x, all_geoms))


def close_small_holes(
    gdf: GeoDataFrame | GeoSeries,
    max_area: int | float,
    *,
    ignore_islands: bool = False,
    copy: bool = True,
) -> GeoDataFrame | GeoSeries:
    """Closes holes in polygons if the area is less than the given maximum.

    It takes a GeoDataFrame or GeoSeries of polygons and
    fills the holes that are smaller than the specified area given in units of
    either square meters ('max_m2') or square kilometers ('max_km2').

    Args:
        gdf: GeoDataFrame or GeoSeries of polygons.
        max_area: The maximum area in the unit of the GeoDataFrame's crs.
        ignore_islands: If False (default), polygons inside the holes (islands)
            will be erased from the "hole" geometries before the area is calculated.
            If True, the entire polygon interiors will be considered, meaning there
            might be duplicate surfaces in the resulting geometries.
            Note that ignoring islands is a lot faster.
        copy: if True (default), the input GeoDataFrame or GeoSeries is copied.
            Defaults to True.

    Returns:
        A GeoDataFrame or GeoSeries of polygons with closed holes in the geometry
        column.

    Raises:
        ValueError: If the coordinate reference system of the GeoDataFrame is not in
            meter units.
        ValueError: If both 'max_m2' and 'max_km2' is given.

    Examples
    --------

    Let's create a circle with a hole in it.

    >>> point = sg.to_gdf([260000, 6650000], crs=25833)
    >>> point
                            geometry
    0  POINT (260000.000 6650000.000)
    >>> circle = sg.buff(point, 1000)
    >>> small_circle = sg.buff(point, 500)
    >>> circle_with_hole = circle.overlay(small_circle, how="difference")
    >>> circle_with_hole.area
    0    2.355807e+06
    dtype: float64

    Close holes smaller than 1 square kilometer (1 million square meters).

    >>> holes_closed = sg.close_small_holes(circle_with_hole, max_area=1_000_000)
    >>> holes_closed.area
    0    3.141076e+06
    dtype: float64

    The hole will not be closed if it is larger.

    >>> holes_closed = sg.close_small_holes(circle_with_hole, max_area=1_000)
    >>> holes_closed.area
    0    2.355807e+06
    dtype: float64
    """
    if not isinstance(gdf, (GeoSeries, GeoDataFrame)):
        raise ValueError(
            f"'gdf' should be of type GeoDataFrame or GeoSeries. Got {type(gdf)}"
        )

    if not len(gdf):
        return gdf

    if copy:
        gdf = gdf.copy()

    gdf = make_all_singlepart(gdf)

    if not ignore_islands:
        all_geoms = make_valid(gdf.unary_union)

        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = gdf.geometry.map(
                lambda x: _close_small_holes_no_islands(x, max_area, all_geoms)
            )
            return gdf
        else:
            return gdf.map(
                lambda x: _close_small_holes_no_islands(x, max_area, all_geoms)
            )
    else:
        geoms = (
            gdf.geometry.to_numpy() if isinstance(gdf, GeoDataFrame) else gdf.to_numpy()
        )
        exteriors = get_exterior_ring(geoms)
        assert len(exteriors) == len(geoms)

        max_rings = max(get_num_interior_rings(geoms))

        if not max_rings:
            return gdf

        # looping through max for all geoms since arrays must be equal length
        interiors = np.array(
            [[get_interior_ring(geom, i) for i in range(max_rings)] for geom in geoms]
        )
        assert interiors.shape == (len(geoms), max_rings), interiors.shape

        areas = area(polygons(interiors))
        interiors[(areas < max_area) | np.isnan(areas)] = None

        results = polygons(exteriors, interiors)

        if isinstance(gdf, GeoDataFrame):
            gdf.geometry = results
            return gdf
        else:
            return GeoSeries(results, crs=gdf.crs)


def _close_small_holes_no_islands(poly, max_area, all_geoms):
    """Closes small holes within one shapely geometry of polygons."""

    # start with a list containing the polygon,
    # then append all holes smaller than 'max_km2' to the list.
    holes_closed = [poly]
    singlepart = get_parts(poly)
    for part in singlepart:
        n_interior_rings = get_num_interior_rings(part)

        if not (n_interior_rings):
            continue

        for n in range(n_interior_rings):
            hole = polygons(get_interior_ring(part, n))
            try:
                no_islands = unary_union(hole.difference(all_geoms))
            except GEOSException:
                no_islands = make_valid(unary_union(hole.difference(all_geoms)))

            if area(no_islands) < max_area:
                holes_closed.append(no_islands)

    return make_valid(unary_union(holes_closed))


def _close_all_holes_no_islands(poly, all_geoms):
    """Closes all holes within one shapely geometry of polygons."""

    # start with a list containing the polygon,
    # then append all holes smaller than 'max_km2' to the list.
    holes_closed = [poly]
    singlepart = get_parts(poly)
    for part in singlepart:
        n_interior_rings = get_num_interior_rings(part)

        for n in range(n_interior_rings):
            hole = polygons(get_interior_ring(part, n))
            try:
                no_islands = unary_union(hole.difference(all_geoms))
            except GEOSException:
                no_islands = make_valid(unary_union(hole.difference(all_geoms)))

            holes_closed.append(no_islands)

    return make_valid(unary_union(holes_closed))


def get_gaps(
    gdf: GeoDataFrame,
    include_interiors: bool = False,
    grid_size: float | int | None = None,
) -> GeoDataFrame:
    """Get the gaps between polygons.

    Args:
        gdf: GeoDataFrame of polygons.
        include_interiors: If False (default), the holes inside individual polygons
            will not be included as gaps.

    Note:
        See get_holes to find holes inside singlepart polygons.

    Returns:
        GeoDataFrame of polygons with only a geometry column.
    """
    if not len(gdf):
        return GeoDataFrame({"geometry": []}, crs=gdf.crs)

    if not include_interiors:
        gdf = close_all_holes(gdf)

    bbox = GeoDataFrame(
        {"geometry": [box(*tuple(gdf.total_bounds)).buffer(1)]}, crs=gdf.crs
    )

    bbox_diff = make_all_singlepart(
        clean_overlay(
            bbox, gdf, how="difference", geom_type="polygon", grid_size=grid_size
        )
    )

    # remove the outer "gap", i.e. the surrounding area
    bbox_ring = get_exterior_ring(bbox.geometry.values)
    without_outer_ring = sfilter_inverse(bbox_diff, bbox_ring)
    return without_outer_ring.reset_index(drop=True)


def get_holes(gdf: GeoDataFrame, as_polygons=True) -> GeoDataFrame:
    """Get the holes inside polygons.

    Args:
        gdf: GeoDataFrame of polygons.
        as_polygons: If True (default), the holes will be returned as polygons.
            If False, they will be returned as LinearRings.

    Note:
        See get_gaps to find holes/gaps between undissolved polygons.

    Returns:
        GeoDataFrame of polygons or linearrings with only a geometry column.
    """
    if not len(gdf):
        return GeoDataFrame({"geometry": []}, index=gdf.index, crs=gdf.crs)

    def as_linearring(x):
        return x

    astype = polygons if as_polygons else as_linearring

    geoms = make_all_singlepart(gdf.geometry).to_numpy()

    rings = [
        GeoSeries(astype(get_interior_ring(geoms, i)), crs=gdf.crs)
        for i in range(max(get_num_interior_rings(geoms)))
    ]

    return (
        GeoDataFrame({"geometry": (pd.concat(rings).pipe(clean_geoms).sort_index())})
        if rings
        else GeoDataFrame({"geometry": []}, crs=gdf.crs)
    )



from typing import Any, Callable, Iterable

import geopandas as gpd
import igraph
import networkx as nx
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from geopandas.array import GeometryArray
from IPython.display import display
from networkx.algorithms import approximation as approx
from numpy import ndarray
from numpy.typing import NDArray
from pandas import Index
from shapely import (
    Geometry,
    STRtree,
    area,
    box,
    buffer,
    centroid,
    difference,
    distance,
    extract_unique_points,
    get_coordinates,
    get_exterior_ring,
    get_interior_ring,
    get_num_interior_rings,
    get_parts,
    get_rings,
    intersection,
    intersects,
    is_empty,
    is_ring,
    length,
    line_merge,
    linearrings,
    linestrings,
    make_valid,
    points,
    polygons,
    segmentize,
    simplify,
    unary_union,
    voronoi_polygons,
)
from shapely.errors import GEOSException
from shapely.geometry import (
    LinearRing,
    LineString,
    MultiLineString,
    MultiPoint,
    Point,
    Polygon,
)

from .conversion import to_gdf, to_geoseries


class PolygonsAsRings:
    """Convert polygons to linearrings, apply linestring functions, then convert back to polygons."""

    def __init__(self, polys: GeoDataFrame | GeoSeries | GeometryArray, crs=None):
        if not isinstance(polys, (pd.DataFrame, pd.Series, GeometryArray)):
            raise TypeError(type(polys))

        self.polyclass = polys.__class__

        if not isinstance(polys, pd.DataFrame):
            polys = to_gdf(polys, crs)

        self._index_mapper = dict(enumerate(polys.index))
        self.gdf = polys.reset_index(drop=True)

        if crs is not None:
            self.crs = crs
        elif hasattr(polys, "crs"):
            self.crs = polys.crs
        else:
            self.crs = None

        if not len(self.gdf):
            self.rings = pd.Series()
            return

        exterior = pd.Series(
            get_exterior_ring(self.gdf.geometry.values),
            index=self._exterior_index,
        )

        self.max_rings: int = np.max(get_num_interior_rings(self.gdf.geometry.values))

        if not self.max_rings:
            self.rings = exterior
            return

        # series same length as number of potential inner rings
        interiors = pd.Series(
            (
                [
                    [get_interior_ring(geom, i) for i in range(self.max_rings)]
                    for geom in self.gdf.geometry
                ]
            ),
        ).explode()

        interiors.index = self._interiors_index

        interiors = interiors.dropna()

        self.rings = pd.concat([exterior, interiors])

    def get_rings(self, agg: bool = False):
        gdf = self.gdf.copy()
        rings = self.rings.copy()
        if agg:
            gdf.geometry = rings.groupby(level=1).agg(unary_union)
        else:
            rings.index = rings.index.get_level_values(1)
            rings.name = "geometry"
            gdf = gdf.drop(columns="geometry").join(rings)

        if issubclass(self.polyclass, pd.DataFrame):
            return GeoDataFrame(gdf, crs=self.crs)
        if issubclass(self.polyclass, pd.Series):
            return GeoSeries(gdf.geometry)
        return self.polyclass(gdf.geometry.values)

    def apply_numpy_func_to_interiors(
        self, func: Callable, args: tuple | None = None, kwargs: dict | None = None
    ):
        """Run an array function on only the interior rings of the polygons."""
        kwargs = kwargs or {}
        args = args or ()
        arr: NDArray[LinearRing] = self.rings.loc[self.is_interior].values
        index: pd.Index = self.rings.loc[self.is_interior].index
        results = pd.Series(
            np.array(func(arr, *args, **kwargs)),
            index=index,
        )
        self.rings.loc[self.is_interior] = results
        return self

    def apply_numpy_func(
        self, func: Callable, args: tuple | None = None, kwargs: dict | None = None
    ):
        """Run a function that takes an array of lines/rings and returns an array of lines/rings."""
        kwargs = kwargs or {}
        args = args or ()

        self.rings.loc[:] = np.array(func(self.rings.values, *args, **kwargs))
        return self

    def apply_geoseries_func(
        self, func: Callable, args: tuple | None = None, kwargs: dict | None = None
    ):
        """Run a function that takes a GeoSeries and returns a GeoSeries."""
        kwargs = kwargs or {}
        args = args or ()

        self.rings.loc[:] = np.array(
            func(
                GeoSeries(
                    self.rings.values,
                    crs=self.crs,
                    index=self.rings.index.get_level_values(1).map(self._index_mapper),
                ),
                *args,
                **kwargs,
            )
        )

        return self

    def apply_gdf_func(
        self, func: Callable, args: tuple | None = None, kwargs: dict | None = None
    ):
        """Run a function that takes a GeoDataFrame and returns a GeoDataFrame."""
        kwargs = kwargs or {}
        args = args or ()

        gdf = GeoDataFrame(
            {"geometry": self.rings.values},
            crs=self.crs,
            index=self.rings.index.get_level_values(1).map(self._index_mapper),
        ).join(self.gdf.drop(columns="geometry"))

        assert len(gdf) == len(self.rings)

        gdf.index = self.rings.index

        self.rings.loc[:] = func(
            gdf,
            *args,
            **kwargs,
        ).geometry.values

        return self

    @property
    def is_interior(self):
        return self.rings.index.get_level_values(0) == 1

    @property
    def is_exterior(self):
        return self.rings.index.get_level_values(0) == 0

    @property
    def _interiors_index(self):
        """A three-leveled MultiIndex.

        Used to separate interior and exterior and sort the interior in
        the 'to_numpy' method.

        level 0: all 1s, indicating "is interior".
        level 1: gdf index repeated *self.max_rings* times.
        level 2: interior number index. 0 * len(gdf), 1 * len(gdf), 2 * len(gdf)...
        """
        if not self.max_rings:
            return pd.MultiIndex()
        len_gdf = len(self.gdf)
        n_potential_interiors = len_gdf * self.max_rings
        gdf_index = sorted(list(self.gdf.index) * self.max_rings)
        interior_number_index = np.tile(np.arange(self.max_rings), len_gdf)
        one_for_interior = np.repeat(1, n_potential_interiors)

        return pd.MultiIndex.from_arrays(
            [one_for_interior, gdf_index, interior_number_index]
        )

    @property
    def _exterior_index(self):
        """A three-leveled MultiIndex.

        Used to separate interior and exterior in the 'to_numpy' method.
        Only leve 1 is used for the exterior.

        level 0: all 0s, indicating "not interior".
        level 1: gdf index.
        level 2: All 0s.
        """
        zero_for_not_interior = np.repeat(0, len(self.gdf))
        return pd.MultiIndex.from_arrays(
            [zero_for_not_interior, self.gdf.index, zero_for_not_interior]
        )

    def to_gdf(self) -> GeoDataFrame:
        """Return the GeoDataFrame with polygons."""
        self.gdf.geometry = self.to_numpy()
        return self.gdf

    def to_geoseries(self) -> GeoDataFrame:
        """Return the GeoDataFrame with polygons."""
        self.gdf.geometry = self.to_numpy()
        return self.gdf.geometry

    def to_numpy(self) -> NDArray[Polygon]:
        """Return a numpy array of polygons."""
        if not len(self.rings):
            return np.array([])

        exterior = self.rings.loc[self.is_exterior].sort_index()
        assert exterior.shape == (len(self.gdf),)

        nonempty_interiors = self.rings.loc[self.is_interior]

        if not len(nonempty_interiors):
            try:
                return make_valid(polygons(exterior.values))
            except Exception:
                return _geoms_to_linearrings_fallback(exterior)

        empty_interiors = pd.Series(
            [None for _ in range(len(self.gdf) * self.max_rings)],
            index=self._interiors_index,
        ).loc[lambda x: ~x.index.isin(nonempty_interiors.index)]

        interiors = (
            pd.concat([nonempty_interiors, empty_interiors])
            .sort_index()
            # make each ring level a column with same length and order as gdf
            .unstack(level=2)
            .sort_index()
        )
        assert interiors.shape == (len(self.gdf), self.max_rings), interiors.shape

        try:
            return make_valid(polygons(exterior.values, interiors.values))
        except Exception:
            return _geoms_to_linearrings_fallback(exterior, interiors)


def get_linearring_series(geoms: Any) -> pd.Series:
    geoms = to_geoseries(geoms).explode(index_parts=False)
    coords, indices = get_coordinates(geoms, return_index=True)
    return pd.Series(linearrings(coords, indices=indices), index=geoms.index)


def _geoms_to_linearrings_fallback(
    exterior: pd.Series, interiors: pd.Series | None = None
) -> pd.Series:
    exterior.index = exterior.index.get_level_values(1)

    exterior = get_linearring_series(exterior)

    if interiors is None:
        return (
            pd.Series(
                make_valid(polygons(exterior.values)),
                index=exterior.index,
            )
            .groupby(level=0)
            .agg(unary_union)
        )

    interiors.index = interiors.index.get_level_values(1)
    new_interiors = []
    for col in interiors:
        new_interiors.append(get_linearring_series(interiors[col]))

    all_none = [[None] * len(new_interiors)] * len(exterior)
    cols = list(interiors.columns)
    out_interiors = pd.DataFrame(
        all_none,
        columns=cols,
        index=exterior.index,
    )
    out_interiors[cols] = pd.concat(new_interiors, axis=1)
    for col in out_interiors:
        out_interiors.loc[out_interiors[col].isna(), col] = None

    return (
        pd.Series(
            make_valid(polygons(exterior.values, out_interiors.values)),
            index=exterior.index,
        )
        .groupby(level=0)
        .agg(unary_union)
    )
